from collections import defaultdict

from antelope import check_direction, comp_dir
from .fragments import LcFragment, DependentFragment


def create_fragment(flow, direction, parent=None, **kwargs):
    """
    User-facing pass-through function. Identical to internal function except that reference fragment direction is
    reversed for UX purposes- to give direction with respect to newly created reference fragment (internal direction
    is always given with respect to parent)
    :param flow:
    :param direction:
    :param parent:
    :param kwargs: uuid, name, comment, value, balance=False, units, ...
    :return:
    """
    if parent is None:
        return _create_fragment(flow, comp_dir(direction), parent=parent, **kwargs)
    else:
        return _create_fragment(flow, direction, parent=parent, **kwargs)


def _create_fragment(flow, direction, uuid=None, parent=None, name=None, comment=None, value=None, balance=False,
                     exchange_value=None,
                     **kwargs):
    """

    :param flow:
    :param direction:
    :param uuid:
    :param parent:
    :param name:
    :param comment:
    :param value:
    :param balance:
    :param kwargs:
    :return:
    """
    direction = check_direction(direction)
    if name is None:
        name = flow['Name']
    name = kwargs.pop('Name', name)

    # handle 'value' vs 'exchange_value' redundancy
    if value is None:
        if exchange_value is not None:
            value = exchange_value
        # counter-case- pass
    else:
        if exchange_value is not None:
            raise TypeError('Cannot provide both "value" and "exchange_value" inputs')
        # counter-case- pass

    if comment is None:
        comment = ''
    comment = kwargs.pop('Comment', comment)
    if parent is None:
        if value is None:
            value = 1.0
        if uuid is None:
            frag = LcFragment.new(name, flow, direction, Comment=comment, exchange_value=value,
                                  **kwargs)
        else:
            frag = LcFragment(uuid, flow, direction, Comment=comment, exchange_value=value, Name=name,
                              **kwargs)
    else:
        # if parent.term.is_null:
        #     parent.to_foreground()  # this is now accomplished inside fragment constructor, via set_parent
        # else:
        #     parent.unset_background()
        if balance or parent.term.is_subfrag:
            # exchange value set during traversal
            value = None

        if uuid is None:
            frag = LcFragment.new(name, flow, direction, parent=parent, Comment=comment,
                                  exchange_value=value, balance_flow=balance, **kwargs)
        else:
            frag = LcFragment(uuid, flow, direction, parent=parent, Comment=comment, exchange_value=value,
                              balance_flow=balance, Name=name, **kwargs)

        # traverse -- may not need to do this anymore if we switch to live traversals for everything
        # parent.traverse(None)  # in fact, let's skip it

    # this cannot be done internally-- really we need create_fragment_from_exchange to do this
    # if flow.context.elementary:
    #     frag.terminate(flow.context)

    return frag


def _transfer_evs(frag, new):
    if frag.observed_ev != 0 and new.observable():
        new.observed_ev = frag.observed_ev
    for scen in frag.exchange_values():
        if scen != 0 and scen != 1 and new.observable(scen):
            new.set_exchange_value(scen, frag.exchange_value(scen))


def clone_fragment(frag, tag='copy', comment=None, _parent=None, origin=None):
    """
    Creates duplicates of the fragment and its children. returns the new reference fragment.
    :param frag:
    :param _parent: used internally
    :param tag: attached to top level fragment
    :param origin:
    :param comment: can be used in place of source fragment's comment
    :return:
    """
    if origin is None:
        origin = frag.origin
    the_comment = comment or frag['Comment']
    the_name = frag['Name']
    if _parent is None:
        the_name += ' (%s)' % tag

    new = _create_fragment(parent=_parent, origin=origin,
                           Name=the_name, StageName=frag.get('StageName'),
                           flow=frag.flow, direction=frag.direction, comment=the_comment,
                           value=frag.cached_ev, balance=frag.is_balance)

    if frag.external_ref != frag.uuid and len(str(tag)) > 0:
        new.external_ref = frag.external_ref + '_%s' % tag

    _transfer_evs(frag, new)  # this obviates the need for observation

    for t_scen, term in frag.terminations():
        if term.term_node is frag or term.is_null:
            new.terminate(None)
        else:
            new.terminate(term.term_node, term_flow=term.term_flow, descend=term.descend,
                          scenario=t_scen)

    for c in frag.child_flows:
        clone_fragment(c, _parent=new, tag=tag, origin=origin)
    return new


def _fork_fragment(fragment, comment=None):
    """
    create a new fragment between the given fragment and its parent. if flow is None, uses the same flow.
    direction is the same as the given fragment. exchange value is shifted to new frag; given frag's ev is
    set to 1.

    The newly created fragment is not terminated and the original fragment's parent is not corrected.  Both of
    these must be done by the calling function before the fragment will have a sensible topology.

    :param fragment:
    :param comment:
    :return: the new fragment
    """
    old_parent = fragment.reference_entity
    subfrag = _create_fragment(parent=old_parent, flow=fragment.flow, direction=fragment.direction,
                               comment=comment, value=fragment.cached_ev,
                               balance=fragment.balance_flow, origin=fragment.origin)
    _transfer_evs(fragment, subfrag)
    fragment.clear_evs()
    return subfrag


def interpose(fragment):
    """
    Insert a new foreground node in-line between the fragment and its parent, terminating it to the foreground and
    making the specified fragment a child flow.
    given fragment sets the new frag as its parent.
    """
    interp = _fork_fragment(fragment, comment='Interposed node')

    # interp.to_foreground()
    fragment.set_parent(interp)  # set_parent calls parent.add_child() which includes parent.to_foreground()

    return interp


def split_subfragment(fragment):
    """
    This method is like interpose except a new reference fragment is created.  The new node becomes a
    cutoff w/r/t its parent, and then gets terminated to the new reference fragment as a subfragment.

    Exchange value and balancing status stays with parent.

    :param fragment:
    :return:
    """
    if fragment.reference_entity is None:
        raise AttributeError('Fragment is already a reference fragment')

    surrogate = _fork_fragment(fragment, comment='New subfragment')

    fragment.unset_parent()
    surrogate.terminate(fragment)

    return fragment



'''
class FragmentEditor(EntityEditor):
    def create_fragment(self, flow, direction, uuid=None, parent=None, comment=None, value=None, balance=False,
                        units=None, **kwargs):
        """

        :param parent:
        :param flow: required. flow or catalog_ref
        :param direction: required. 'Input' or 'Output'
          (note that direction reversal for reference fragments happens internally!)
        :param uuid: [None]
        :param comment:
        :param value:
        :param balance:
        :param units: [None] if both value and unit are non-None, interpret value as given in units and convert
        :param kwargs:
        :return:
        """
        self._print('Creating fragment with flow %s' % flow)
        if direction not in directions:
            direction = None
        if self._interactive:
            direction = direction or {'i': 'Input', 'o': 'Output'}[cyoa('flow is (I)nput or (O)utput?', 'IO').lower()]
        if direction is None:
            raise ValueError('Must supply direction')
        if 'Name' in kwargs:
            name = kwargs.pop('Name')
        else:
            name = flow['Name']

        # TODO: add me to ContextRefactor
        if value is not None and units is not None:
            value *= flow.reference_entity.convert(units)

        comment = comment or self.ifinput('Enter FragmentFlow comment: ', '')
        if parent is None:
            # direction reversed for UX! user inputs direction w.r.t. fragment, not w.r.t. parent
            if value is None:
                value = 1.0

            if uuid is None:
                frag = LcFragment.new(name, flow, comp_dir(direction), Comment=comment, exchange_value=value,
                                      **kwargs)
            else:
                frag = LcFragment(uuid, flow, comp_dir(direction), Comment=comment, exchange_value=value, Name=name,
                                  **kwargs)
        else:
            if parent.term.is_null:
                parent.to_foreground()
            if balance or parent.term.is_subfrag:
                if self._interactive:
                    print('Exchange value set during traversal')
                value = 1.0
            else:
                if value is None:
                    val = self.ifinput('Exchange value (%s per %s): ' % (flow.unit(), parent.unit), '1.0')
                    if val == '1.0':
                        value = 1.0
                    else:
                        value = parse_math(val)

            if uuid is None:
                frag = LcFragment.new(name, flow, direction, parent=parent, Comment=comment,
                                      exchange_value=value, balance_flow=balance, **kwargs)
            else:
                frag = LcFragment(uuid, flow, direction, parent=parent, Comment=comment, exchange_value=value,
                                  balance_flow=balance, Name=name, **kwargs)

        """ # can't hack this without a qdb
        if self._qdb.is_elementary(frag.flow):
            frag.terminate(frag.flow)
        """
        return frag

    @staticmethod
    def _update_ev(frag, scenario):
        val = frag.exchange_value(scenario)
        cur = str(val)
        upd = ifinput('New value: ', cur)
        if upd != cur:
            new_val = parse_math(upd)
            if new_val != val:
                if scenario is None:
                    frag.reset_cache()
                    frag.cached_ev = new_val
                else:
                    frag.set_exchange_value(scenario, new_val)

    def revise_exchanges(self, frag, scenario=None):
        """
        interactively update reference exchange values (or values for a particular scenario) for the children of
        a given fragment
        :param frag:
        :param scenario:
        :return:
        """
        print('Reference flow: [%s] %s' % (comp_dir(frag.direction), frag.unit))
        if scenario is None:
            print('Update reference flow')
        else:
            print('Update reference flow for scenario "%s"' % scenario)
        self._update_ev(frag, scenario)
        for c in frag.child_flows:
            print('   Child flow: %s ' % c)
            if scenario is None:
                print('Update default value')
            else:
                print('Update value for scenario "%s"' % scenario)
            self._update_ev(c, scenario)

    @staticmethod
    def transfer_evs(frag, new):
        if frag.observed_ev != 0 and new.observable():
            new.observed_ev = frag.observed_ev
        for scen in frag.exchange_values():
            if scen != 0 and scen != 1 and new.observable(scen):
                new.set_exchange_value(scen, frag.exchange_value(scen))

    def clone_fragment(self, frag, suffix=' (copy)', comment=None, origin=None, _parent=None):
        """
        Creates duplicates of the fragment and its children. returns the new reference fragment.
        :param frag:
        :param _parent: used internally
        :param suffix: attached to top level fragment
        :param comment: can be used in place of source fragment's comment
        :param origin: [None] defaults to frag.origin
        :return:
        """
        if _parent is None:
            direction = comp_dir(frag.direction)  # this gets re-reversed in create_fragment
        else:
            direction = frag.direction
        if suffix is None:
            suffix = ''
        if origin is None:
            origin = frag.origin
        the_comment = comment or frag['Comment']
        new = self.create_fragment(parent=_parent, origin=origin,
                                   Name=frag['Name'] + suffix, StageName=frag['StageName'],
                                   flow=frag.flow, direction=direction, comment=the_comment,
                                   value=frag.cached_ev, balance=frag.is_balance,
                                   background=frag.is_background)

        self.transfer_evs(frag, new)

        for t_scen, term in frag.terminations():
            if term.is_null:
                continue
            elif term.term_node is frag:
                new.to_foreground(scenario=t_scen)
            else:
                new.terminate(term.term_node, term_flow=term.term_flow, direction=term.direction,
                              descend=term.descend, inbound_ev=term.inbound_exchange_value,
                              scenario=t_scen)

        for c in frag.child_flows:
            self.clone_fragment(c, _parent=new, suffix='')
        return new

    def _fork_fragment(self, fragment, comment=None):
        """
        create a new fragment between the given fragment and its parent. if flow is None, uses the same flow.
        direction is the same as the given fragment. exchange value is shifted to new frag; given frag's ev is
        set to 1.

        The newly created fragment is not terminated and the original fragment's parent is not corrected.  Both of
        these must be done by the calling function before the fragment will have a sensible topology.

        :param fragment:
        :param comment:
        :return: the new fragment
        """
        old_parent = fragment.reference_entity
        subfrag = self.create_fragment(parent=old_parent, flow=fragment.flow, direction=fragment.direction,
                                       comment=comment, value=fragment.cached_ev,
                                       balance=fragment.is_balance)
        self.transfer_evs(fragment, subfrag)
        fragment.clear_evs()
        return subfrag

    def interpose(self, fragment):
        """
        Insert a new foreground node in-line between the fragment and its parent, terminating it to the foreground and
        making the specified fragment a child flow.
        given fragment sets the new frag as its parent.
        """
        interp = self._fork_fragment(fragment, comment='Interposed node')

        interp.to_foreground()
        fragment.set_parent(interp)

        return interp

    def split_subfragment(self, fragment):
        """
        This method is like interpose except a new reference fragment is created.  The new node becomes a
        cutoff w/r/t its parent, and then gets terminated to the new reference fragment as a subfragment.

        Exchange value and balancing status stays with parent.

        :param fragment:
        :return:
        """
        surrogate = self._fork_fragment(fragment, comment='New subfragment')

        fragment.unset_parent()
        surrogate.terminate(fragment)

        return fragment
<<<<<<< HEAD:antelope_catalog/foreground/fragment_editor.py
'''


def set_child_exchanges(fragment, scenario=None, reset_cache=False):
    """
    This is really client code, doesn't use any private capabilities
    Set exchange values of child flows based on inventory data for the given scenario.  The termination must be
     a foreground process.

    In order for this function to work, flows in the node's exchanges have to have the SAME external_ref as the
    child flows, though origins can differ.  There is no other way for the exchanges to be set automatically from
    the inventory.  Requiring that the flows have the same name, CAS number, compartment, etc. is too fragile /
    arbitrary.  The external refs must match.

    This works out okay for databases that use a consistent set of flows internally -- ILCD, thinkstep, and
    ecoinvent all seem to have that characteristic but ask me again in a year.

    To automatically set child exchanges for different scenarios that use processes from different databases,
    encapsulate each term node inside a sub-fragment, and then specify different subfragment terminations for the
    different scenarios.  Then, map each input / output in the sub-fragment to the correct foreground flow using
    a conserving child flow.

    In that case, the exchange values will be set during traversal, and each sub-fragment's internal exchange
    values can be set automatically using set_child_exchanges.

    :param fragment:
    :param scenario: [None] for the default scenario, set observed ev
    :param reset_cache: [False] if True, for the default scenario set cached ev
    :return:
    """
    term = fragment.termination(scenario)
    if not term.term_node.entity_type == 'process':
        raise DependentFragment('Child flows are set during traversal')

    if scenario is None and fragment.reference_entity is None:
        # this counts as observing the reference flow
        if fragment.observed_ev == 0:
            fragment.observed_ev = fragment.cached_ev

    children = defaultdict(list)  # need to allow for differently-terminated child flows -- distinguish by term.id

    for k in fragment.child_flows:
        key = (k.flow.external_ref, k.direction)
        children[key].append(k)
    if len(children) == 0:
        return

    for x in term.term_node.inventory(ref_flow=term.term_flow):
        if x.value is None:
            fragment.dbg_print('skipping None-valued exchange: %s' % x)
            continue

        key = (x.flow.external_ref, x.direction)
        if key in children:
            try:
                if len(children[key]) > 1:
                    child = next(c for c in children[key] if c.termination(scenario).id == x.termination)
                else:
                    child = next(c for c in children[key])
            except StopIteration:
                continue

            fragment.dbg_print('setting %s [%10.3g]' % (child, x.value))
            if scenario is None:
                if reset_cache:
                    child.reset_cache()
                    child.cached_ev = x.value
                else:
                    child.observed_ev = x.value
            else:
                child.set_exchange_value(scenario, x.value)



from antelope import UnknownOrigin
from antelope_core.archives import InterfaceError
from antelope_core.catalog import LcCatalog
from .foreground_query import ForegroundQuery, ForegroundNotSafe, MissingResource

from itertools import chain
import shutil
import os
import re
import logging


foreground_origin_regexp = re.compile('^[A-Za-z0-9_-]+(\.[A-Za-z0-9_-]+)*$')
savefile_regexp = re.compile('^([A-Za-z0-9_-]+(\.[A-Za-z0-9_-]+)*)\.(\d+)\.(\d+)\.zip$')


class BackReference(Exception):
    """
    trying to instantiate a foreground that's currently being loaded
    """
    pass


class NoSuchForeground(Exception):
    """
    foregrounds must be explicitly created
    """
    pass


class OriginDependencies(object):
    """
    This class stores records of references made from one origin (host) to another (dependency), detected
    whenever an LcForeground
    """
    def __init__(self):
        self._deps = set()  #

    @property
    def all(self):
        for d in sorted(self._deps):
            yield d

    def add_dependency(self, host, dependency):
        self._deps.add((str(host), str(dependency)))

    def dependencies(self, host):
        for h, d in sorted(self._deps):
            if h == host:
                yield d

    def recursive_dependencies(self, *hosts):
        """
        we want all the downstream dependencies but without repeating or entering an endless loop
        :param hosts:
        :return:
        """
        hosts_seen = set()
        deps_seen = set()
        queue = list(hosts)
        while len(queue) > 0:  # this sort of construction always gives me a thrill
            h = queue.pop(0)
            if h in hosts_seen:
                continue
            hosts_seen.add(h)
            for d in self.dependencies(h):
                queue.append(d)
                if d not in deps_seen:
                    deps_seen.add(d)
                    yield d

    def referents(self, dependency):
        for h, d in sorted(self._deps):
            if d == dependency:
                yield h


class ForegroundCatalog(LcCatalog):
    """
    Adds the ability to create (and manage?) foreground resources

    Maintains two different lists of resources that have been encountered but not yet resolved:

     _fg_queue is a set of foregrounds (that may reference one another)- they can be added to the queue
      when referenced, and their queries will resolve once the queue is processed, which will happen within
      a single query evaluation

     _missing_o is a set of (origin, interface) 2-tuples that have been requested but are not found. They must
      be removed from the set when a resource is added to fulfill them.
    """

    '''
    ForegroundCatalog
    '''
    def __init__(self, *args, **kwargs):
        self._fg_queue = set()  # fgs we are *currently* opening
        self._missing_o = set()  # references we have encountered that we cannot resolve
        self._dependencies = OriginDependencies()
        super(ForegroundCatalog, self).__init__(*args, **kwargs)

    @property
    def dependencies(self):
        for d in self._dependencies.all:
            yield d

    def _check_missing_o(self, res):
        for iface in res.interfaces:
            key = (res.origin, iface)
            if key in self._missing_o:
                self._missing_o.remove(key)

    def new_resource(self, reference, source, ds_type, store=True, **kwargs):
        res = super(ForegroundCatalog, self).new_resource(reference, source, ds_type, store=store, **kwargs)
        self._check_missing_o(res)
        return res

    def add_resource(self, resource, **kwargs):
        super(ForegroundCatalog, self).add_resource(resource, **kwargs)
        self._check_missing_o(resource)

    def is_in_queue(self, home):
        """
        This tells us whether the foreground named in home is actively being instantiated
        :param home:
        :return:
        """
        return home in self._fg_queue

    def is_missing(self, origin, interface):
        """
        This tells us whether the named origin, interface tuple is in our list of missing references
        :param origin:
        :param interface:
        :return:
        """
        return origin, interface in self._missing_o

    @property
    def missing_resources(self):
        for k in self._missing_o:
            yield k

    '''
    def delete_foreground(self, ref):
        """
        
        :param ref: 
        :return: 
        """
        """
        Creates or activates a foreground as a sub-folder within the catalog's root directory.  Returns a
        Foreground interface.
        :param path: either an absolute path or a subdirectory path relative to the catalog root
        :param ref: semantic reference (optional)
        :param quiet: passed to fg archive
        :param reset: [False] if True, clear the archive and create it from scratch, before returning the interface
        :param delete: [False] if True, delete the existing tree completely and irreversibly. actually just rename
        the directory to whatever-DELETED; but if this gets overwritten, there's no going back.  Overrides reset.
        :return:
        if localpath:
            if os.path.exists(localpath):
                del_path = localpath + '-DELETED'
                if os.path.exists(del_path):
                    rmtree(del_path)
                os.rename(abs_path, del_path)
        dels = [k for k in self._resolver.resolve(ref, interfaces='foreground')]
        """
        dels = [k for k in self._resolver.resolve(ref, interfaces='foreground')]
        for k in dels:
            self.delete_resource(k, delete_source=True, delete_cache=True)
    '''

    def gen_interfaces(self, origin, itype=None, strict=False, ):
        """
        Override parent method to also create local backgrounds
        :param origin:
        :param itype:
        :param strict:
        :return:
        """
        if origin in self.foregrounds:
            for res in self._sorted_resources(origin, itype, strict):
                '''
                self._fg_queue.add(origin)
                res.check(self)
                self._fg_queue.remove(origin)
                '''
                try:
                    self._check_foreground(res)
                    yield res.make_interface(itype)
                except InterfaceError:
                    continue

        elif (origin, itype) in self._missing_o:
            if itype == 'quantity':
                # try locally first
                for k in super(ForegroundCatalog, self).gen_interfaces(self._qdb.ref, itype=itype):
                    yield k
            raise MissingResource(origin, itype)

        else:
            try:
                for k in super(ForegroundCatalog, self).gen_interfaces(origin, itype=itype, strict=strict):
                    yield k
            except (UnknownOrigin, InterfaceError):
                if itype == 'quantity':
                    # try locally first
                    for k in super(ForegroundCatalog, self).gen_interfaces(self._qdb.ref, itype=itype):
                        yield k
                self._missing_o.add((origin, itype))
                raise MissingResource(origin, itype)

    @property
    def test(self):
        return bool(self._test)

    def create_foreground(self, ref, path=None, quiet=True):
        """
        Creates foreground resource and returns an interface to that resource.
        By default creates in a subdirectory of the catalog root with the ref as the folder
        :param ref:
        :param path:
        :param quiet:
        :return:
        """
        if ref in self._nicknames:
            ref, _ = self._nicknames[ref]

        if ref in self.foregrounds:
            raise KeyError('Foreground %s already exists' % ref)

        assert bool(foreground_origin_regexp.match(ref)), "Foreground reference not valid: %s" % ref
        if self._test:
            if path is not None and os.path.exists(path):
                local_path = path
            else:
                local_path = ref
        else:
            if path is None:
                path = os.path.join(self._rootdir, ref)  # should really sanitize this somehow
                # localpath = ref
            else:
                if os.path.isabs(path):
                    pass
                    # localpath = None
                else:
                    # localpath = path
                    path = os.path.join(self._rootdir, path)

            abs_path = os.path.abspath(path)
            local_path = self._localize_source(abs_path)

        res = self.new_resource(ref, local_path, 'LcForeground',
                                interfaces=['basic', 'index', 'foreground', 'quantity'],
                                quiet=quiet)

        return self._check_foreground(res)

    def foreground(self, ref, reset=False, create=False):
        """
        activates a foreground resource and returns a query interface to that resource.

        :param ref:
        :param reset: re-load the foreground from the saved files
        :param create: [True] run create_foreground(ref) or [False] raise NoSuchForeground
        :return:
        """
        if ref in self._nicknames:
            ref, _ = self._nicknames[ref]

        if ref in self._fg_queue:
            raise BackReference(ref)

        try:
            res = next(self._resolver.resolve(ref, interfaces='foreground'))
        except (UnknownOrigin, StopIteration):
            if create:
                return self.create_foreground(ref)
            else:
                raise NoSuchForeground(ref)

        if reset:
            self.purge_resource_archive(res)

        return self._check_foreground(res)

    def _get_target_abspath(self, target_dir=None):
        if target_dir is None:
            target_dir = self.root
        else:
            target_dir = os.path.abspath(target_dir)
            if not os.path.isdir(target_dir):
                raise NotADirectoryError(target_dir)
        return target_dir

    def _saved_versions(self, foreground, target_dir=None):
        """
        Generates a list of versioned foreground savefiles for the named foreground, as produced by write_versioned_fg,
        in REVERSE ORDER (most recent first), sorted by the numeric values of the major and minor version numbers.

        This routine hard-codes the filename convention of (foreground origin).(major version).(minor version).zip
        and conventional usage assumes all files can be found in the catalog's root directory
        :param foreground:
        :param target_dir: specify where to look for saved versions (default cat.root)
        :return:
        """
        def _make_sortable(filename):
            """
            here we have to take something that says "my.origin.[majorversion].[minorversion]" and make it so that
            [majorversion].10 appears after [majorversion].1 IN GENERAL

            so we pad using %d.%04d and we prohibit more than 9,999 minor versions

            :param filename:
            :return:
            """
            g = savefile_regexp.match(filename).groups()
            return float('%d.%04d' % (int(g[-2]), int(g[-1])))

        candidates = [file for file in sorted(os.listdir(self._get_target_abspath(target_dir))) if
                      bool(savefile_regexp.match(file)) and savefile_regexp.match(file).groups()[0] == foreground]

        for save in sorted(candidates, key=lambda x: _make_sortable(x), reverse=True):
            yield save

    @staticmethod
    def _restore_foreground_zip_to(source_file, target):
        print('Unpacking lastsave %s to %s' % (source_file, target))
        os.makedirs(target)
        shutil.unpack_archive(source_file, extract_dir=target)

    def restore_foreground_by_version(self, foreground, major_version, minor_version=0, target_dir=None, force=False):
        target_dir = self._get_target_abspath(target_dir)
        ref = '%s.%d.%d' % (foreground, major_version, minor_version)
        file = os.path.join(target_dir, '%s.zip' % ref)
        try:
            res = self.get_resource(ref)
            target = res.source
        except UnknownOrigin:
            target = os.path.join(target_dir, ref)
        if os.path.exists(file):
            if os.path.exists(target):
                if force:
                    if os.path.isdir(target_dir):
                        shutil.rmtree(target)
                    else:
                        os.remove(target)
                else:
                    raise FileExistsError(target)
            self._restore_foreground_zip_to(file, target)
            return self.create_foreground(ref, path=target)
        else:
            raise FileNotFoundError(file)

    def _check_foreground(self, res):
        """
        finish foreground activation + return QUERY interface
        If the foreground source directory doesn't exist, BUT at least one versioned save file DOES exist, the highest-
        versioned save file is expanded and renamed to the foreground source directory.
        :param res:
        :return:
        """
        ref = res.origin

        abs_source = self.abs_path(res.source)

        if not os.path.exists(abs_source):
            try:
                lastsave = next(self._saved_versions(ref))
                savedir, ext = os.path.splitext(lastsave)
                assert ext == '.zip'
                assert savedir.startswith(ref)

                source_file = os.path.join(self._rootdir, lastsave)

                self._restore_foreground_zip_to(source_file, abs_source)

            except StopIteration:
                # no save files
                pass

        self._fg_queue.add(ref)
        res.check(self)
        self._fg_queue.remove(ref)

        if ref not in self._queries:
            self._seed_fg_query(ref)
            self.get_archive(ref, strict=True).make_interface('foreground')  # finish the job

        return self._queries[ref]

    def flush_backreference(self, origin):
        if origin in self._fg_queue:
            self._fg_queue.remove(origin)

    @property
    def foregrounds(self):
        f = set()
        for k in self.interfaces:
            org, inf = k.split(':')
            if inf == 'foreground' and org not in f:
                yield org
                f.add(org)

    def clear_unit_scores(self, lcia_method):
        logging.info('Clearing unit scores for %s' % lcia_method.link)
        for f in self.foregrounds:
            if f in self._queries:
                self.get_archive(f).clear_unit_scores(lcia_method)

    def write_versioned_fg(self, foreground, target_dir=None, force=False):
        target_dir = self._get_target_abspath(target_dir)

        ar = self.get_archive(foreground)
        new_ref = '.'.join([foreground, str(ar.metadata.version_major), str(ar.metadata.version_minor)])
        new_path = os.path.join(target_dir, new_ref)

        if force:
            if os.path.isdir(new_path):
                shutil.rmtree(new_path)
        else:
            fname = new_path + '.zip'
            if os.path.exists(fname):
                raise FileExistsError(fname)
            if os.path.isdir(new_path):
                raise IsADirectoryError(new_path)

        shutil.copytree(ar.source, new_path)
        zipfile = shutil.make_archive(new_path, format='zip', root_dir=new_path)
        shutil.rmtree(new_path)
        return zipfile

    '''
    def assign_new_origin(self, old_org, new_org):
        """
        This only works for certain types of archives. Foregrounds, in particular. but it is hard to say what else.
        What needs to happen here is:
         - first we retrieve the archive for the ref (ALL archives?)
         - then we call set_origin() on the archive
         - then we save the archive
         - then we rename the resource file
         = actually we just rewrite the resource file, since the filename and JSON key have to match
         = since we can't update resource origins, it's easiest to just blow them away and reload them
         = but to save time we should transfer the archives from the old resource to the new resource
         = anyway, it's not clear when we would want to enable this operation in the first place.
         * so for now we leave it
        :param old_org:
        :param new_org:
        :return:
        """
        pass

    def configure_resource(self, reference, config, *args):
        """
        We must propagate configurations to internal, derived resources. This also begs for testing.
        :param reference:
        :param config:
        :param args:
        :return:
        """
        # TODO: testing??
        for res in self._resolver.resolve(reference, strict=False):
            abs_src = self.abs_path(res.source)
            if res.add_config(config, *args):
                if res.internal:
                    if os.path.dirname(abs_src) == self._index_dir:
                        print('Saving updated index %s' % abs_src)
                        res.archive.write_to_file(abs_src, gzip=True,
                                                  exchanges=False, characterizations=False, values=False)
                else:
                    print('Saving resource configuration for %s' % res.origin)
                    res.save(self)

            else:
                if res.internal:
                    print('Deleting unconfigurable internal resource for %s\nsource: %s' % (res.origin, abs_src))
                    self.delete_resource(res, delete_source=True)
                else:
                    print('Unable to apply configuration to resource for %s\nsource: %s' % (res.origin, res.source))

    def delete_foreground(self, foreground, really=False):
        res = self.get_resource(foreground, 'foreground')
        self.delete_resource(res)
        abs_src = self.abs_path(res.source)
        if really:
            shutil.rmtree(abs_src)
        else:
            del_path = abs_src + '-DELETED'
            if os.path.exists(del_path):
                shutil.rmtree(del_path)

            os.rename(abs_src, del_path)
    '''

    def _seed_fg_query(self, origin, **kwargs):
        self._queries[origin] = ForegroundQuery(origin, catalog=self, **kwargs)

    def query(self, origin, strict=False, refresh=False, **kwargs):
        if origin in self.foregrounds:
            if origin not in self._queries:
                # we haven't loaded this fg yet, so
                raise ForegroundNotSafe(origin)
            if origin in self._fg_queue:
                raise BackReference(origin)
            if refresh or (origin not in self._queries):
                self._seed_fg_query(origin, **kwargs)
            return self._queries[origin]

        return super(ForegroundCatalog, self).query(origin, strict=strict, refresh=refresh, **kwargs)

    def internal_ref(self, fg_ref, origin, external_ref):
        """
        for use by an LcForeground provider to obtain a reference to an entity from a separate origin
        :param fg_ref: referring foreground ref
        :param origin:
        :param external_ref:
        :return:
        """
        self._dependencies.add_dependency(fg_ref, origin)
        try:
            return self.query(origin).get(external_ref)
        except UnknownOrigin:
            self._missing_o.add((origin, 'basic'))
            raise MissingResource(origin, 'basic')
        # don't catch EntityNotFound

    '''
    Parameterization
    Observing fragments requires the catalog because observations can come from different resources.
    '''



"""
Flow Terminations are model-defined links between a particular flow and a process that terminates it.

They originated as part of LcFragments but in fact they are more general. A FlowTermination is actually the same
as a ProductFlow in lca-matrix, plus features to compute LCIA.  It should be easy to construct either one from the
other.
"""
import logging

from antelope import (BackgroundRequired, check_direction, comp_dir, QuantityRequired, MultipleReferences,
                      NoReference, ConversionReferenceMismatch, EntityNotFound, NoCatalog)

from antelope.models import DirectedFlow
from antelope_core.contexts import NullContext
from antelope_core.exchanges import ExchangeValue
from antelope_core.lcia_results import LciaResult
from antelope_core.implementations.quantity import do_lcia
from .lcia_dict import LciaResults
from .models import Anchor, EntityRef, UNRESOLVED_ANCHOR_TYPE


# from lcatools.catalog_ref import NoCatalog
# from lcatools.interact import parse_math


class MissingFlow(Exception):
    """
    Raised when a termination does not match with the specified term_flow
    """
    pass


class FlowConversionError(Exception):
    pass


class SubFragmentAggregation(Exception):
    pass


class UnresolvedAnchor(Exception):
    pass


class NonConfigurableInboundEV(Exception):
    """
    only foreground terminations may have their inbound exchange values explicitly specified (and this via reference
    flow observation, NOT through specifying it on the anchor)
    """
    pass


class UnCachedScore(Exception):
    """
    means that we have an LCIA-only node whose score has not been set for the requested LCIA method
    """
    pass


class TerminationFromJson(Exception):
    """
    Something failed in deserialization
    """
    pass


class MalformedParent(Exception):
    """
    parent's flow or direction is None
    """
    pass


class FlowTermination(object):

    _term = None
    _term_flow = None
    _direction = None
    _descend = True

    """
    these are stored by scenario in a dict on the mainland

    A fragment can have the following types of terminations:
     * None - the termination is null- the flow enters the foreground and becomes an i/o
     * parent - the fragment's termination is the fragment itself.  The fragment flow  enters a foreground node.
       The node can have children but only has LCIA impacts based on the terminating flow, which have to be looked up
       in the database. fg-terminated nodes don't have scenarios (e.g. the scenarios are in the exchange values).
       Note: term_flows can be different from parent flows, and unit conversions will occur normally (e.g. "sulfur
       content" converted to "kg SO2")
     * Process - the flow enters a process referenced by CatalogRef.  The node's LCIA impacts are fg_lcia. The
       node's children are the process's non-term intermediate exchanges. The node can also have other children.
       (created with terminate or term_from_exch)
     * Fragment - the flow enters a sub-fragment.  The sub-fragment must be traversable. The node's children are
       the fragment's non-term io flows. The node cannot have other children.  If the sub-fragment is background,
       then the background fragment flow supplants the foreground one during traversal.

    LCIA results are always cached in the terminations, and are not (by default) persistent across instantiations.
    """
    @classmethod
    def from_json(cls, fragment, fg, scenario, j):
        if len(j) == 0:
            return cls.null(fragment)
        origin = j.pop('source', None) or j.pop('origin')
        if origin == 'foreground':
            origin = fg.ref

        # handle term flow
        tf_ref = j.pop('termFlow', None)
        if tf_ref is None:
            term_flow = None  # if it is unspecified, let the best value be determined
        elif isinstance(tf_ref, dict):
            term_flow = fg.catalog_ref(tf_ref['origin'], tf_ref['externalId'], entity_type='flow')
        else:
            if origin == fg.ref:
                term_flow = fg[tf_ref]
            else:
                term_flow = fg.catalog_ref(origin, tf_ref, entity_type='flow')

        if 'context' in j:
            term_node = fg.tm[j['context']]
        else:
            try:
                external_ref = j['externalId']
            except KeyError:
                external_ref = j['entityId']
            # handle term_node
            if origin == fg.ref:
                term_node = fg[external_ref]
            else:
                term_node = fg.catalog_ref(origin, external_ref, entity_type=UNRESOLVED_ANCHOR_TYPE)  # need to lookup type

        direction = j.pop('direction', None)
        descend = j.pop('descend', True)
        try:
            term = cls(fragment, term_node, _direction=direction, term_flow=term_flow, descend=descend)
        except MultipleReferences:
            raise TerminationFromJson('term_flow missing and ambiguous: %s (%s)' % (fragment.link, scenario))
        except FlowConversionError:
            t = term_flow or term_node
            raise TerminationFromJson('Flow Conversion Error %s =/= %s' % (fragment.link, t.link))
        if 'scoreCache' in j.keys():
            term._deserialize_score_cache(fg, j['scoreCache'], scenario)
        return term

    '''
    @classmethod
    def from_exchange_ref(cls, fragment, exchange_ref):
        return cls(fragment, exchange_ref.process_ref, direction=exchange_ref.direction,
                   term_flow=exchange_ref.exchange.flow, inbound_ev=exchange_ref.exchange.value)

    @classmethod
    def from_term(cls, fragment, term):
        return cls(fragment, term.term_node, direction=term.direction, term_flow=term.term_flow,
                   descend=term.descend, inbound_ev=term.inbound_exchange_value)
    '''

    @classmethod
    def null(cls, fragment):
        return cls(fragment, None)

    def __init__(self, fragment, entity, _direction=None, term_flow=None, descend=True, inbound_ev=None):
        """
        reference can be None, an entity or a catalog_ref.  It only must have origin, external_ref, and entity_type.
        To use an exchange, use FlowTermination.from_exchange()
         * None - to create a foreground IO / cutoff flow
         * fragment (same as parent) - to create a foreground node.  Must satisfy 'fragment is entity'
         * process or process ref - to link the fragment to a process inventory (uses self.is_background to determine
           foreground or background lookup)
         * context - to represent the fragment's flow (or term_flow, still supported) as emission
         * flow or flow ref - no longer supported.  Supply context instead.

        The term's direction is detected at the time of creation.

        :param fragment:
        :param entity:
        :param _direction: Should not be used externally- only if term is generated from_json
        :param term_flow: optional flow to match on termination inventory or LCIA.  If flow and term_flow have
        different reference quantities, quantity conversion is performed during traversal
        :param descend:
        :param inbound_ev: ignored; deprecated
        """
        if fragment.flow is None or fragment.direction is None:
            raise MalformedParent(fragment.link)
        self._parent = fragment
        if entity is not None:
            if entity.entity_type == 'flow':
                if term_flow is not None and term_flow != entity:
                    raise ValueError('Inconsistent flow and term_flow provided: %s' % entity)
                term_flow = entity
                entity = fragment  # foreground termination with flow conversion
                # raise TypeError('Can no longer terminate fragments with flows. Use context instead')
            elif entity.entity_type == 'exchange':
                term_flow = entity.flow
                _direction = entity.direction
                entity = entity.process
            elif entity.entity_type not in ('context', 'process', 'fragment', UNRESOLVED_ANCHOR_TYPE):
                raise TypeError('%s: term %s Inappropriate termination type: %s' % (fragment.link, entity.link,
                                                                                    entity.entity_type))

            # check for recursive loops
            if entity.entity_type == 'fragment' and entity is not fragment:
                if entity.top() is fragment.top():
                    # interior recursive loop can be resolved by leaving cut-off

                    print('-- setting cut-off flow to resolve recursive loop')
                    entity = None

        self._term = entity  # this must have origin, external_ref, and entity_type, and be operable (if ref)
        self._score_cache = LciaResults(fragment)

        self.term_flow = term_flow
        self.direction = _direction
        self.descend = descend

    @property
    def term_node(self):
        if self._term:
            if self._term.entity_type == UNRESOLVED_ANCHOR_TYPE:
                try:
                    self._term = self._term.resolve()
                except NoCatalog:
                    pass
        return self._term

    @property
    def term_flow(self):
        if self._term_flow is None:
            if self.is_process:
                raise AttributeError('[%s] term_flow was not specified for process term!' % self._parent.external_ref)
            if self.is_frag:
                return self.term_node.flow
            else:
                return self._parent.flow
        return self._term_flow

    @term_flow.setter
    def term_flow(self, term_flow):
        """
        Introduce term validation checking here if needed
        :param term_flow:
        :return:
        """
        if term_flow is None:
            if self.is_process:
                try:
                    self._term_flow = self.term_node.reference().flow
                except MultipleReferences as e:
                    try:
                        self._term_flow = self.term_node.reference(self._parent.flow).flow
                    except KeyError:
                        raise e

            elif self.is_frag:
                self._term_flow = None  # leave unspecified to plug into term's ref flow
            else:
                self._term_flow = self._parent.flow
        else:
            if self.is_process:
                try:
                    self._term_flow = self.term_node.reference(term_flow).flow
                except NoReference:  # we need to allow processes with no reference to be used as term nodes
                    self._term_flow = term_flow
                except KeyError:
                    raise MissingFlow(term_flow)
            elif self.is_frag:
                if term_flow == self.term_node.flow:  # don't need the inventory every time
                    self._term_flow = term_flow
                elif term_flow in (x.flow for x in self.term_node.cutoffs()):
                    self._term_flow = term_flow
                else:
                    raise MissingFlow(term_flow)
            else:
                self._term_flow = term_flow
        # this was causing us problems on delayed queries with MissingResource and it doesn't add anything
        # if self.valid and self.node_weight_multiplier == 0:  # we don't need to re-validate the flow
        #     print('Warning: 0 node weight multiplier for term of %s' % self._parent.external_ref)

    @property
    def direction(self):
        if self.is_process:
            return self._direction
        return comp_dir(self._parent.direction)

    @property
    def valid(self):
        if self.is_null:
            return True  # cutoff fragment flows should not report zero node weight
        return self.term_node.validate()

    @direction.setter
    def direction(self, value):
        if value is None:
            # this is the default: should set the direction by the reference.  Only non-none if from_json
            if self.is_process:
                rx = self.term_node.reference(self.term_flow)
                self._direction = rx.direction
                # for fg, invert direction doesn't make sense. for subfragments, direction is ignored
        else:
            self._direction = check_direction(value)

    '''
    def matches(self, exchange):
        """
        returns True if the exchange specifies the same process and flow as the term's process_ref and term_flow
        :param exchange:
        :return:
        """
        if self.is_null:
            return False
        if self.term_node.entity_type != 'process':
            return False
        return (self._term.external_ref == exchange.process.external_ref) and (self.term_flow.match(exchange.flow))

    def terminates(self, exchange):
        """
        Returns True if the exchange's termination matches the term's term_node, and the flows also match, and the
        directions are complementary.
        If the exchange does not specify a termination, returns True if the flows match and directions are comp.
        :param exchange:
        :return:
        """
        if self.term_flow.match(exchange.flow) and self.direction == comp_dir(exchange.direction):
            if exchange.termination is None:
                return True
            else:
                if self.is_null:
                    return False
                if self.term_node.entity_type != 'process':
                    return False
                if exchange.termination == self._term.external_ref:
                    return True
        return False

    def to_exchange(self):
        if self.is_null:
            return None
        return ExchangeValues(self.term_node, self.term_flow, self.direction, value=self.inbound_exchange_value)
    '''

    @property
    def is_local(self):
        """
        Fragment and termination have the same origin. Implies is_frag, since processes cannot be added to foregrounds
        :return:
        """
        if self.is_null:
            return False
        return self._parent.origin == self.term_node.origin

    @property
    def is_context(self):
        """
        termination is a context
        :return:
        """
        if self.is_null:
            return False
        return self.term_node.entity_type == 'context'

    @property
    def is_frag(self):
        """
        Termination is a fragment
        :return:
        """
        return (not self.is_null) and (self.term_node.entity_type == 'fragment')

    @property
    def is_process(self):
        """
        termination is a process
        :return:
        """
        return (not self.is_null) and (self.term_node.entity_type == 'process') and self.valid

    @property
    def is_emission(self):
        """
        Pending context refactor
        :return:
        """
        return self.is_context and self.term_node.elementary

    @property
    def is_fg(self):
        """
        Termination is parent
        :return:
        """
        return (not self.is_null) and (self.term_node is self._parent)

    @property
    def is_bg(self):
        """
        A term is "background" if the traversal does not include any descending fragments.  This condition is
        met in either of the following two cases:
         - the parent is background (no child flows) and the termination is to a process (i.e. not fg or cutoff)
         - the termination is to a fragment and descend is False
        old:
        # parent is marked background, or termination is a background fragment
        :return:
        """
        return (self._parent.is_background and self.is_process) or (self.is_subfrag and not self.descend)

    @property
    def term_is_bg(self):
        """
        Termination is local and background.  This is deprecated and always returns False.

        (Reason for its existence dates to when 'background' was a user-specified designation, and fragments were
        often terminated to a singleton-background fragment (fragment with an immediate background node and no children)

        The idea was simply to shorten the list of FragmentFlows generated in a traversal by removing the "connective
        tissue" fragment that terminated to the singleton background.  Now that background status is automatically
        determined by number of children, this is not reliable.)
        :return:
        """
        return False
        # return self.is_frag and self.is_local and self.term_node.is_background

    @property
    def is_subfrag(self):
        """
        Termination is a non-self fragment.  (we were excluding background frags too but that is outmoded)

        Old: Controversy around whether expression should be:
        self.is_frag and not (self.is_fg or self.is_bg or self.term_is_bg)  [current] or
        self.is_frag and (not self.is_fg) and (not self.is_bg)  [old; seems wrong]

        :return:
        """
        return self.is_frag and not self.is_fg  # or self.is_bg or self.term_is_bg)

    @property
    def is_null(self):
        return self._term is None

    @property
    def is_unresolved(self):
        return self._term is not None and self._term.entity_type == UNRESOLVED_ANCHOR_TYPE

    @property
    def descend(self):
        return self._descend

    @descend.setter
    def descend(self, value):
        if value is None:
            return
        if isinstance(value, bool):
            self._descend = value
            ''' # this whole section not needed- we can certainly cache LCIA scores for nondescend fragments,
            and we don't need to blow them away if descend is True; just ignore them.
            if value is True:
                self.clear_score_cache()  # if it's descend, it should have no score_cache
                # if it's not descend, the score gets computed (and not cached) during traversal
            '''
        else:
            raise ValueError('Descend setting must be True or False')

    @property
    def name(self):
        if self.is_null:
            name = self.term_flow['Name']
        elif self.is_context:
            name = '%s, %s' % (self.term_flow['Name'], self.term_node.name)
        else:
            name = self.term_node.name
        return name

    @property
    def term_ref(self):
        if self.is_null:
            return None
        elif self.is_context:
            return self.term_node.name
        return self.term_node.external_ref

    @property
    def flow_conversion(self):
        """
        express the parent's flow in terms of the quantity of the term flow.
        There are two ways to do this, each case involving the quantity relation on either the parent flow or the
        term flow, between the two quantities (parent flow's r.q. is the reference quantity; term flow's r.q. is the
        query quantity).

        In each case, we want the flow's native term manager to perform the conversion using ITS OWN canonical
        quantities.  The assumption is that the parent flow's r.q. is tied to our local LciaEngine, while the
        term flow's r.q. could be either local or remote.

        The QuantityRef.quantity_relation() implicitly assumes that the invoking quantity is the QUERY quantity, so
        the "forward" (natural parent->node) direction uses the remote flow's r.q. - but we do the "reverse" direction
        first because it's local.

        how to deal with scenario cfs? tbd
        problem is, the term doesn't know its own scenario

        :return: float = amount in term_flow ref qty that corresponds to a unit of fragment flow's ref qty
        """
        '''
        What are all the possibilities?
         the parent quantity knows a characterization for the parent flow w.r.t. the term quantity
         the parent quantity knows a characterization for the term flow w.r.t. the term quantity
         the term quantity knows a characterization for the parent flow w.r.t. the parent quantity
         the term quantity knows a characterization for the term flow w.r.t. the parent quantity

        "hey look at me, net calorific value. see that mass flow f4- a unit of it is worth 35 of me"
        is how the database is constructed.
        
        but each call to quantity_relation should check for both forward and reverse matches
        '''
        if not self.valid:
            logging.warning('Flow Conversion attempted on invalid term node %5.5s' % self._parent.uuid)
            return 1.0
        # if not self.term_flow.validate():
        #     return 1.0
        if self.term_flow is None:
            raise MissingFlow(self)
        if self._parent.flow is None:
            raise MalformedParent(self._parent)
        if self.term_flow.reference_entity == self._parent.flow.reference_entity:
            return 1.0
        parent_q = self._parent.flow.reference_entity
        term_q = self.term_flow.reference_entity
        if term_q is None:
            return 0.0

        # first - natural - ask our parent flow if fit can convert to term quantity
        try:
            rev = self._parent.flow.cf(term_q, context=self.term_node.external_ref)
            if rev == 0.0:
                rev = self._parent.flow.cf(term_q)

        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            rev = None

        if rev:
            return rev

        # then ask the term_flow if it can convert to parent quantity
        try:
            fwd = self.term_flow.cf(parent_q)
        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            fwd = None

        if fwd:
            return 1.0 / fwd

        # then ask if our parent qty can convert term_flow
        try:
            # I have no idea what unwritten rules are governing this context declaration - 2024-07-01
            rev_c = parent_q.quantity_relation(self.term_flow, term_q, context=(self.term_node.origin,
                                                                                self.term_node.external_ref))
        except (QuantityRequired, NotImplementedError):
            rev_c = None
        except ConversionReferenceMismatch:
            try:
                rev_c = parent_q.quantity_relation(self.term_flow, term_q, context=None)
            except ConversionReferenceMismatch:
                rev_c = None

        if rev_c:
            return 1.0 / rev_c.value

        # last, ask if remote quantity recognizes *our* flow
        print(' %s: flow_conversion: untested' % self._parent.link)
        try:
            fwd_c = term_q.quantity_relation(self._parent.flow, parent_q, context=(self.term_node.origin,
                                                                                   self.term_node.external_ref))
        except (QuantityRequired, NotImplementedError, ConversionReferenceMismatch):
            fwd_c = None

        if fwd_c:
            print('reverse q-hit')
            return fwd_c.value
        print(' %s: flow_conversion FAILED' % self._parent.link)
        raise FlowConversionError('Zero CF found relating %s to %s' % (self.term_flow, self._parent.flow))

    @property
    def id(self):
        if self.is_null:
            return None
        else:
            return self.term_node.external_ref

    @property
    def inbound_exchange_value(self):
        """
        This is only used for correcting fragment-term direction mismatches.
        This needs to be tested!
        :return:
        """
        if self.direction == self._parent.direction:
            return -1.0
        return 1.0

    @inbound_exchange_value.setter
    def inbound_exchange_value(self, val):
        raise NonConfigurableInboundEV

    @property
    def node_weight_multiplier(self):
        return self.flow_conversion / self.inbound_exchange_value

    @property
    def unit(self):
        if self.is_null:
            return '--'
        if self.term_node.entity_type == 'fragment':  # fg, bg, or subfragment
            return '%4g unit' % self.inbound_exchange_value
        return '%4g %s' % (self.inbound_exchange_value, self.term_flow.unit)  # process

    @property
    def observed_flows(self):
        for cf in self._parent.child_flows:
            yield DirectedFlow.from_observed(cf)

    def unobserved_exchanges(self, refresh=False):
        """
        Generator which yields exchanges from the term node's inventory that are not found among the child flows, for
          LCIA purposes

        Challenge here going forward: we made some kind of normative decision early on that terminations do not know
        their own scenarios, that the fragment maps scenario to termination. The problem is that now terminations
        cannot themselves carry out traversal on the term_node because they don't know what scenarios to pass.

        The upshot is that we cannot meaningfully compute "unobserved exchanges" for subfragments, since we don't
        know our scenario.

        :return:
        """
        if self.is_context:
            x = ExchangeValue(self._parent, self.term_flow, self._parent.direction, termination=self.term_node,
                              value=self.node_weight_multiplier)  # TODO: need to figure out how we are handling locales
            yield x
        elif self.is_frag:  # fragments can have unobserved exchanges too! (CAN THEY?)
            # this code is unreachable because self.is_frag returns a null LciaResult in compute_unit_score()
            for x in []:
                yield x
        else:
            if self.is_bg:  # or len(list(self._parent.child_flows)) == 0:
                # ok we're bringing it back but only because it is efficient to cache lci
                for x in self.term_node.lci(ref_flow=self.term_flow, refresh=refresh):
                    yield x
            else:
                for x in self.term_node.unobserved_lci(self.observed_flows, ref_flow=self.term_flow):
                    yield x  # this should forward out any cutoff exchanges

    def _fallback_lcia(self, quantity_ref, locale, **kwargs):
        """
        Uses term node inventory if no background is available
        :param quantity_ref:
        :param locale:
        :param kwargs:
        :return:
        """
        print('WARNING- BackgroundRequired - using fallback lcia %s' % self._parent.link)
        child_flows = set((k.flow.external_ref, k.direction) for k in self._parent.child_flows)
        inv = [x for x in self.term_node.inventory(ref_flow=self.term_flow)
               if (x.flow.external_ref, x.direction) not in child_flows]
        res = quantity_ref.do_lcia(inv, locale=locale, **kwargs)
        return res

    def compute_unit_score(self, quantity_ref, refresh=False, **kwargs):
        """
        four different ways to do this.
        0- we are a subfragment-- no direct impacts unless non-descend, which is caught earlier
        1- parent is bg: ask catalog to give us bg_lcia (process or fragment)
        2- get fg lcia for unobserved exchanges

        If
        :param quantity_ref:
        :param refresh:
        :return:
        """
        if self.is_context:
            try:
                locale = self._parent['SpatialScope']
            except KeyError:
                locale = self.term_flow.locale
            x = ExchangeValue(self._parent, self.term_flow, self._parent.direction, termination=self.term_node,
                              value=self.node_weight_multiplier)  # TODO: need to figure out how we are handling locales
            res = do_lcia(quantity_ref, [x], locale=locale, refresh=refresh, **kwargs)

        else:
            # OK, so we are not frag and we are not context and we are not null-- we are process!
            try:
                locale = self.term_node['SpatialScope']
            except KeyError:
                locale = 'GLO'

            try:
                res = self.term_node.bg_lcia(quantity_ref, observed=self.observed_flows, ref_flow=self.term_flow,
                                             refresh=refresh, locale=locale, **kwargs)
            except (QuantityRequired, EntityNotFound, NotImplementedError):
                try:
                    res = quantity_ref.do_lcia(self.unobserved_exchanges(refresh=refresh), locale=locale,
                                               refresh=refresh, **kwargs)
                except (BackgroundRequired, NotImplementedError):
                    res = self._fallback_lcia(quantity_ref, locale, **kwargs)
            except BackgroundRequired:
                res = self._fallback_lcia(quantity_ref, locale, **kwargs)

        if isinstance(res, list):
            [k.scale_result(self.inbound_exchange_value) for k in res]
        else:
            res.scale_result(self.inbound_exchange_value)
        return res

    def score_cache(self, quantity=None, refresh=False, **kwargs):
        """
        only process-terminations are cached
        remote fragments that come back via the API can have cached scores as well, but local subfragments should not
        get cached.

        :param quantity:
        :param ignore_uncached:
        :param refresh: If True, re-compute unit score even if it is already present in the cache.
        :param kwargs:
        :return:
        """
        if quantity is None:
            return self._score_cache

        if self.is_null or self.is_fg:
            return LciaResult(quantity)

        if not self.valid:
            raise UnresolvedAnchor

        if refresh:
            self._score_cache.pop(quantity, None)

        if quantity in self._score_cache:
            return self._score_cache[quantity]
        else:
            if self.is_frag:  # but not fg, ergo subfrag
                raise UnCachedScore(quantity)
            else:
                '''
                # This refresh situation is a problem.  On the one hand, if we don't pass refresh on to do_lcia,
                then there's no way to clear "seen cfs" on flow refs.  On the other hand, passing it through recursively
                from frag_flow_lcia means we will continually "see" and then "refresh" seen cfs on every flow, for every
                fragment we traverse. This is just an efficiency bomb.  We need to find a better way to refresh scores
                that won't suffer from this problem- but unfortunately I can't think of anyway to track that state 
                mid-traversal. 
                
                note: additionally, refreshing "seen cfs" and refreshing fragment unit scores are separate
                tasks and should not use the same keyword argument.  In both cases, passing 'refresh' to a recursive
                traversal results in highly duplicated computation.  
                
                The solution is to no longer support 'refresh' during traversal or any recursive operation.  However,
                we retain it here because we still want to be able to do both of:
                 - refresh seen cfs during calls to do_lcia() in compute_unit_score()
                 - refresh cached LCI during calls to lci() in _unobserved_exchanges()
                
                As to the other uses:
                 - individual termination scores can be refreshed using this method score_cache()
                 - entire foreground scores can be refreshed using the clear_unit_scores() implementation method
                 - flow CFs (apart from those included in specified terminations) can only be refreshed by manually 
                    calling do_lcia() with an inventory that includes them.  This is obviously not ideal, and a new 
                    solution should be sought.  
                '''
                res = self.compute_unit_score(quantity, refresh=refresh, **kwargs)
            if isinstance(res, list):
                for k in res:
                    self._score_cache[k.quantity] = k
            else:
                self._score_cache[quantity] = res
            return self._score_cache[quantity]

    def score_cache_items(self):
        return self._score_cache.items()

    def lcia(self):
        for k, v in self.score_cache_items():
            print('%s' % v)

    def reset_score(self, lcia):
        self._score_cache.pop(lcia, None)

    def clear_score_cache(self):
        self._score_cache.clear()

    def _serialize_score_cache(self):
        """
        Score cache contains an LciaResults object, which works as a dict.
        serialization should preserve order, which prohibits using a simple dict
        :return: a list to be serialized directly
        """
        score_cache = []
        for q in self._score_cache.indices():
            res = self._score_cache[q]
            score_cache.append({'quantity': {'origin': res.quantity.origin,
                                             'externalId': res.quantity.external_ref},
                                'score': res.total()})
        return score_cache

    def add_lcia_score(self, quantity, score, scenario=None):
        res = LciaResult(quantity, scenario=scenario)
        res.add_summary(self._parent.external_ref, self._parent, 1.0, score)
        self._score_cache[quantity] = res

    def _deserialize_score_cache(self, fg, sc, scenario):
        self._score_cache = LciaResults(self._parent)
        for i in sc:
            q = fg.catalog_ref(i['quantity']['origin'], i['quantity']['externalId'], entity_type='quantity')
            self.add_lcia_score(q, i['score'], scenario=scenario)

    def _term_flow_block(self):
        if self.term_flow.origin == self.term_node.origin:
            return self.term_flow.external_ref
        else:
            return {
                'origin': self.term_flow.origin,
                'externalId': self.term_flow.external_ref
            }

    def serialize(self, save_unit_scores=False):
        if self.is_null:
            return {}
        if self.is_context:
            j = {
                'origin': 'foreground',
                'context': self.term_node.name
            }
        else:
            j = {
                'origin': self.term_node.origin,
                'externalId': self.term_node.external_ref
            }
            if self.is_local:
                j['origin'] = 'foreground'  # override

        # saving term_flow: for subfragments, we save it only it it's specified
        if self.is_frag:
            if self._term_flow is not None:
                j['termFlow'] = self._term_flow_block()
        elif self.term_flow != self._parent.flow:
            j['termFlow'] = self._term_flow_block()

        if self.direction != comp_dir(self._parent.direction):
            j['direction'] = self.direction
        if self._descend is False:
            j['descend'] = False
        if self._parent.is_background and save_unit_scores and len(self._score_cache) > 0:
            j['scoreCache'] = self._serialize_score_cache()
        return j

    def to_anchor(self, save_unit_scores=False, group=None):
        if self.is_null:
            return Anchor.null()
        d = {'descend': self.descend}
        if self._parent.is_background and save_unit_scores and len(self._score_cache) > 0:
            d['score_cache'] = self._serialize_score_cache()
        if self.is_context:
            d['context'] = self.term_node.as_list()
            if self.term_flow != self._parent.flow:
                d['anchor_flow'] = EntityRef.from_entity(self.term_flow)
            return Anchor(**d)
        else:
            if self.is_frag:
                if self.term_flow != self.term_node.flow:
                    d['anchor_flow'] = EntityRef.from_entity(self.term_flow)
            else:
                if self.term_flow != self._parent.flow:
                    d['anchor_flow'] = EntityRef.from_entity(self.term_flow)
            d['node'] = EntityRef.from_entity(self.term_node)
            return Anchor(**d)

    def __eq__(self, other):
        """
        Terminations are equal if they are both null, both fg, or if term_node, term_flow, direction and descend match
        :param other:
        :return:
        """
        if self is other:
            return True
        if not isinstance(other, FlowTermination):
            return False
        if self.is_null:
            if other.is_null:
                return True
            return False
        if self.is_fg:
            if other.is_fg:
                return True
            return False
        return (self.term_node.external_ref == other.term_node.external_ref and
                self.term_flow == other.term_flow and
                self.direction == other.direction and
                self.descend == other.descend)  # probably want to remove this

    def __str__(self):
        """

        :return:
          '---:' = fragment I/O
          '-O  ' = foreground node
          '-*  ' = process
          '-#  ' - sub-fragment (aggregate)
          '-#::' - sub-fragment (descend)
          '-B ' - terminated background
          '--C ' - cut-off background
          '--? ' - ungrounded catalog ref
        """
        if self.is_null:
            term = '---:'  # fragment IO
        elif self.is_fg:
            term = '-O  '
        elif self.is_context:
            if self.is_emission:
                term = '-== '
            elif self.term_node is NullContext:
                term = '-)  '
            else:
                # TODO: intermediate contexts don't present as cutoffs (because is_null is False)
                term = '-cx '
        elif self.term_node.entity_type == 'process':
            if self.is_bg:
                term = '-B* '
            else:
                term = '-*  '
        elif self.term_node.entity_type == 'fragment':
            if self.descend:
                term = '-#::'
            else:
                term = '-#  '
        elif self.term_node.entity_type == UNRESOLVED_ANCHOR_TYPE:
            term = '--? '
        else:
            raise TypeError('I Do not understand this term for frag %.7s' % self._parent.uuid)
        return term



"""
I forget how to work this

Basically we want to create a foreground with a few interfaces:
"""

from ..foreground_catalog import ForegroundCatalog
from antelope_core.entities.tests.base_testclass import refinery_archive
from ..terminations import FlowConversionError
from antelope_core.characterizations import DuplicateCharacterizationError

import unittest


class TestFlowConversions(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cat = ForegroundCatalog.make_tester()
        cat.new_resource('test.refinery', refinery_archive, 'json', interfaces=('basic', 'index',
                                                                                'exchange', 'quantity',
                                                                                'background'),
                         config={'hints': [['context', 'air', 'to air'],
                                           ['context', 'water', 'to water']]})
        cat.create_foreground('my.main')
        cat.create_foreground('my.alt')

        cat.foreground('my.main').new_flow('Massive flow A', 'mass')

        cls.cat = cat

    @property
    def mass(self):
        return self.cat.get_canonical('mass')

    @property
    def volume(self):
        return self.cat.get_canonical('volume')

    @property
    def gasoline(self):
        return next(self.cat.query('test.refinery').flows(name='gasoline'))

    @property
    def ncv(self):
        return self.cat.get_canonical('net calorific value')

    @property
    def main(self):
        return self.cat.foreground('my.main')

    @property
    def alt(self):
        return self.cat.foreground('my.alt')

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cat.__del__()

    def test_gasoline_flow(self):
        self.assertEqual(self.gasoline.cf(self.mass), 750.0)

    def test_mass_to_gasoline(self):
        """
        This tests the use of the remote qdb to access flow properties
        :return:
        """
        p = next(self.gasoline.targets())
        a = self.main.new_flow('Massive flow A', self.mass)
        f = self.main.new_fragment(a, 'Output', exchange_value=1.0)
        f.terminate(p, term_flow=self.gasoline)
        self.assertEqual(f.term.flow_conversion * 750.0, 1.0)

    def test_anchor_flow(self):
        """
        Just a straight up test of above using fragments
        :return:
        """
        p = next(self.gasoline.targets())
        f_0 = self.main.new_fragment(self.gasoline, 'Output', observe=True, termination=p)
        z = self.alt.new_flow('Massive Flow Z', 'mass')
        f_z = self.alt.new_fragment(z, 'Output', observe=True, termination=f_0)
        self.assertEqual(f_z.term.flow_conversion * 750.0, 1.0)

    def test_volume_to_gasoline(self):
        """
        This tests the identity hit
        :return:
        """
        p = next(self.gasoline.targets())
        b = self.main.new_flow('Volume flow B', self.volume)
        f = self.main.new_fragment(b, 'Output', exchange_value=1.0)
        f.terminate(p, term_flow=self.gasoline)
        self.assertEqual(f.term.flow_conversion, 1.0)

    def test_fg_conv_gasoline(self):
        """
        This tests the use of the local qdb to access ad hoc characterizations of remote flows
        :return:
        """
        p = next(self.gasoline.targets())
        e = self.main.new_flow('Energetic flow E', self.ncv)
        self.gasoline.characterize(e.reference_entity, 32250.0)
        f = self.main.new_fragment(e, 'Output', exchange_value=1.0)
        f.terminate(p, term_flow=self.gasoline)
        self.assertEqual(f.term.flow_conversion * 32250.0, 1.0)

    def test_fg_conv_direct(self):
        """
        This tests the use of local qdb to characterize local flows
        :return:
        """
        p = next(self.gasoline.targets())
        g = self.main.new_flow('Massive fuel flow', self.mass)
        g.characterize(self.volume, 0.00122)
        f = self.main.new_fragment(g, 'Output', exchange_value=1.0)
        f.terminate(p, term_flow=self.gasoline)
        self.assertEqual(f.term.flow_conversion, 0.00122)

    def test_intermediate_context(self):
        """
        This tests whether we can use anchor-specific CFs
        :return:
        """
        e = self.main.new_flow('Energetic flow E', self.ncv)
        f_e = self.main.new_fragment(e, 'Output', observe=True)
        h = self.alt.new_flow('Fuel flow H', 'mass')
        f_h = self.alt.new_fragment(h, 'Output', observe=True)

        f_h.terminate(f_e)
        with self.assertRaises(FlowConversionError):
            f_h.traverse()

        h.characterize(self.ncv, 123.0)

        f_h.traverse()
        self.assertEqual(f_h.term.flow_conversion, 123.0)

        """ # we don't actually raise DuplicateCharacterizationError- we catch it and IGNORE it. possibly unwise.
        with self.assertRaises(DuplicateCharacterizationError):
            h.characterize(self.ncv, 126.0)
        """
        h.characterize(self.ncv, 125.0)
        self.assertEqual(f_h.term.flow_conversion, 123.0)  # test that 125 was ignored

        h.characterize(self.ncv, 126.0, context=(f_e.origin, f_e.external_ref))
        self.assertEqual(f_h.term.flow_conversion, 126.0)  # test that 126 was accepted


if __name__ == '__main__':
    unittest.main()



"""
LCA Test

We need an efficient way to test a model under code changes and evolution to identify when results change.
This architecture was inspired by a set of ad hoc tests I wrote for CATRA-yoy.  The idea is to generate a
concise serializable record of an LCA computation and store it, and then later on to deserialize and compare to live
results.

The serializable record is a set of tuples ('contextual name', 'contextual_value'). For testing purposes, a string
equality combined with a stable rounding method was judged to provide the best confidence.  "rounding" method applies
the format '%.*g' % (self.precision, contextual_value).  precision defaults to 8.

"""

from pydantic import BaseModel, field_serializer, ConfigDict
from typing import Set, Tuple, List
from antelope import QuantityRef, EntityNotFound, CatalogRef
from antelope_core.entities import LcQuantity
from antelope.models import EntityRef

from collections import defaultdict
import json
import os


def _num(entry):
    name, value = entry
    try:
        return float(value)
    except TypeError:
        print('glitch on value %s' % value)
        return 0.0


class _FragmentLcaTest(BaseModel):
    test: str
    """
    A set of results for a given resource.  Subclasses will specify HOW to generate + compare the results
    """
    origin: str
    external_ref: str
    precision: int
    scenarios: List[str] = []
    results: Set[Tuple[str, str]] = {}

    @property
    def name(self):
        return 'Test %s: %s/%s (%s)' % (self.test, self.origin, self.external_ref, self.scenarios)

    @classmethod
    def from_model(cls, model, *scenarios, precision=None, **kwargs):
        if precision is None:
            precision = 8
        cot = cls(origin=model.origin, external_ref=model.external_ref, precision=precision, scenarios=scenarios,
                  **kwargs)
        cot.set(model)
        return cot

    def rounding(self, value):
        return '%.*g' % (self.precision, value)

    def run(self, model) -> Set[Tuple[str, str]]:
        return NotImplemented

    def set(self, model):
        self.results = self.run(model)

    def check(self, model):
        test = self.run(model)
        first = test.difference(self.results)
        second = self.results.difference(test)
        if len(first) == 0 and len(second) == 0:
            return True
        else:
            print('In Test:')
            if first:
                for i in sorted(first, key=_num, reverse=True):
                    print(i)
            else:
                print('all accounted')
            print('\nIn Results:')
            if second:
                for i in sorted(second, key=_num, reverse=True):
                    print(i)
            else:
                print('all accounted')
            return False


class CutoffsTest(_FragmentLcaTest):
    test: str = 'cutoff'

    def run(self, model):
        cos = model.cutoffs(scenario=self.scenarios)
        return set((k.flow.external_ref, self.rounding(k.value)) for k in cos)


class ActivityTest(_FragmentLcaTest):
    """
    Note: this does not work for dynamically generated models because UUIDs won't match- use FragmentFlowsTest instead
    """
    test: str = 'activity'

    def run(self, model):
        act = model.activity(scenario=self.scenarios)
        return set((obj.fragment.external_ref, self.rounding(obj.node_weight)) for obj in act)


class FragmentFlowsTest(_FragmentLcaTest):
    """
    really this result set should be a list, but it will still run deterministically because it will
    store all encountered combinations of each flow + magnitude

    """
    test: str = 'flows'

    def run(self, model):
        ffs = model.traverse(scenario=self.scenarios)
        return set((ff.fragment.flow.name, self.rounding(ff.magnitude)) for ff in ffs)


class _LciaTest(_FragmentLcaTest):
    """
    This is sloppy because we need live quantities to run the tests
    we wouldn't need to do this if we had a query
    """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    lcia_method: QuantityRef | LcQuantity  # we keep a live query on hand

    @property
    def name(self):
        return 'Test %s: %s/%s (%s) | %s' % (self.test, self.origin, self.external_ref, self.scenarios,
                                             self.lcia_method.name)

    def check(self, model):
        if not self.lcia_method.is_entity:
            if not self.lcia_method.resolved:
                return False
        return super(_LciaTest, self).check(model)

    @field_serializer('lcia_method')
    def serialize_field(v):
        """
        we serialize as an EntityRef so that we can retrieve it at runtime
        :return:
        """
        return EntityRef.from_entity(v)


class LciaTotalTest(_LciaTest):
    test: str = 'lcia_total'

    def run(self, model):
        res = model.fragment_lcia(self.lcia_method, scenario=self.scenarios)
        return {(self.external_ref, self.rounding(res.total()))}


class LciaContribTest(_LciaTest):
    test: str = 'lcia_contrib'

    def run(self, model):
        res = model.fragment_lcia(self.lcia_method, scenario=self.scenarios)
        return {(obj.name, self.rounding(obj.cumulative_result)) for obj in res.components()} | \
            {('total', self.rounding(res.total()))}


class LciaAggTest(_LciaTest):
    test: str = 'lcia_agg'

    def _make_entry(self, obj):
        return obj.entity.name, self.rounding(obj.cumulative_result)

    def run(self, model):
        res = model.fragment_lcia(self.lcia_method, scenario=self.scenarios, mode='stage')
        return {self._make_entry(c) for c in res.components()} | \
            {('total', self.rounding(res.total()))}


class LciaFlatTest(LciaAggTest):
    test: str = 'lcia_flat'

    def run(self, model):
        res = model.fragment_lcia(self.lcia_method, scenario=self.scenarios, mode='flat')
        return {self._make_entry(c) for c in res.components()} | \
            {('total', self.rounding(res.total()))}


_my_mapping = {
    'cutoff': CutoffsTest,
    'activity': ActivityTest,
    'flows': FragmentFlowsTest,
    'lcia_total': LciaTotalTest,
    'lcia_contrib': LciaContribTest,
    'lcia_agg': LciaAggTest,
    'lcia_flat': LciaFlatTest
}


class LcaTestSuite(BaseModel):
    cutoff: List[CutoffsTest] = []
    activity: List[ActivityTest] = []
    flows: List[FragmentFlowsTest] = []
    lcia_total: List[LciaTotalTest] = []
    lcia_contrib: List[LciaContribTest] = []
    lcia_agg: List[LciaAggTest] = []
    lcia_flat: List[LciaFlatTest] = []
    
    def _traversal_test(self, test, model, *scenarios, **kwargs):
        obj = _my_mapping[test].from_model(model, *scenarios, **kwargs)
        getattr(self, test).append(obj)
        
    def add_cutoff_test(self, model, *args, precision=None):
        self._traversal_test('cutoff', model, *args, precision=precision)

    def add_activity_test(self, model, *args, precision=None):
        self._traversal_test('activity', model, *args, precision=precision)

    def add_flows_test(self, model, *args, precision=None):
        self._traversal_test('flows', model, *args, precision=precision)

    def add_lcia_total_test(self, model, quantity_ref, *args, **kwargs):
        self._traversal_test('lcia_total', model, *args, lcia_method=quantity_ref, **kwargs)

    def add_lcia_contrib_test(self, model, quantity_ref, *args, **kwargs):
        self._traversal_test('lcia_contrib', model, *args, lcia_method=quantity_ref, **kwargs)

    def add_lcia_agg_test(self, model, quantity_ref, *args, **kwargs):
        self._traversal_test('lcia_agg', model, *args, lcia_method=quantity_ref, **kwargs)

    def add_lcia_flat_test(self, model, quantity_ref, *args, **kwargs):
        self._traversal_test('lcia_flat', model, *args, lcia_method=quantity_ref, **kwargs)

    @classmethod
    def from_tests(cls, *tests):
        spec = defaultdict(list)
        for test in tests:
            spec[test.test].append(test)

        return cls(**spec)

    def save(self, filename, overwrite=False):
        if os.path.exists(filename):
            if overwrite is False:
                raise FileExistsError(filename)
        j = self.model_dump_json(indent=2)
        with open(filename, 'w') as fp:
            fp.write(j)
        print('wrote %d tests to %s' % (len(self), filename))

    def __len__(self):
        return sum(len(getattr(self, s)) for s in _my_mapping.keys())

    def run_tests(self, cat, check_origin=None, apply_scenarios=None):
        results = defaultdict(list)
        ps = ct = sk = 0
        for group in _my_mapping.keys():
            tests = getattr(self, group)
            for test in tests:
                print('\n%s' % test.name)
                if check_origin:
                    org = check_origin
                else:
                    org = test.origin
                try:
                    model = cat.query(org).get(test.external_ref)
                except EntityNotFound:
                    sk += 1
                    print('entity %s/%s not found - skipping' % (org, test.external_ref))
                    continue
                if apply_scenarios:
                    hold = model.scenarios
                    model.scenarios = apply_scenarios
                    o = test.check(model)
                    model.scenarios = hold
                else:
                    o = test.check(model)
                results[group].append(o)
                if o:
                    ps += 1
                ct += 1
        print('Passed %d out of %d tests (%d skipped)' % (ps, ct, sk))


def deserialize_test_suite(cat, j):
    """
    This takes a json-serialized LcaTestSuite as input and converts all the entity_refs into live entities prior
    to deserialization

    :param cat:
    :param j:
    :return:
    """
    for tt, tests in j.items():
        if tt.startswith('lcia_'):
            for test in tests:
                e_ref = test.pop('lcia_method')
                try:
                    q = cat.query(e_ref['origin']).get(e_ref['entity_id'])
                except EntityNotFound:
                    q = CatalogRef.from_json(e_ref)
                test['lcia_method'] = q
    return LcaTestSuite(**j)


def run_tests(cat, testfile, check_origin=None):
    with open(testfile) as fp:
        j = json.load(fp)

    suite = deserialize_test_suite(cat, j)

    suite.run_tests(cat, check_origin=check_origin)
    return suite



"""


"""

import uuid
import logging
# from collections import defaultdict

from antelope import (comp_dir, check_direction, PropertyExists, CatalogRef, RxRef, QuantityRequired,
                      RefQuantityRequired, ConversionError)

from ..fragment_flows import group_ios, FragmentFlow, ios_exchanges, frag_flow_lcia, FragmentInventoryDeprecated
from antelope_core.entities import LcEntity, LcFlow
# from antelope_core.exchanges import ExchangeValue
# from lcatools.interact import ifinput, parse_math
from ..terminations import FlowTermination, MissingFlow
from ..foreground_query import MissingResource, FragmentBranch


class InvalidParentChild(Exception):
    pass


class FoundBalanceFlow(Exception):
    """
    raised if a fragment attempts to traverse a balance flow the normal way
    """
    pass


class BalanceAlreadySet(Exception):
    pass


class CacheAlreadySet(Exception):
    pass


class ScenarioConflict(Exception):
    pass


class DependentFragment(Exception):
    pass


class ZeroInboundExchange(Exception):
    pass


def _new_evs():
    d = dict()  # should be a LowerDict? should scenario names be case sensitive?
    d[0] = 1.0
    d[1] = 0.0
    return d


class LcFragment(LcEntity):
    """

    """

    @classmethod
    def new(cls, name, *args, verbose=False, **kwargs):
        """
        :param name: the name of the fragment
        :param args: need flow and direction
        :param verbose: print when a fragment is created
        :param kwargs: parent, exchange_value, private, balance_flow, background, termination
        :return:
        """
        if verbose:
            print('LcFragment - Name: %s:' % name, 0)  # this will never run since the constructor does not set dbg
        return cls(uuid.uuid4(), *args, Name=name, **kwargs)

    _ref_field = 'parent'

    _new_fields = ['StageName']

    @classmethod
    def from_json(cls, fg, j):
        if j['parent'] is None or j['parent'].lower() == 'none':
            parent = None
        else:
            parent = fg[j['parent']]
            if parent is None:
                print('warning: parent %s returned None' % j['parent'])
        flow_block = j['flow']
        if isinstance(flow_block, dict):  # "modern" but still legacy
            flow = fg.catalog_ref(flow_block['origin'], flow_block['externalId'])
        else:
            flow = fg[j['flow']]
        if flow is None:
            try:
                org, ext = j['flow'].split('/')
                flow = fg.catalog_ref(org, ext, entity_type='flow')
            except ValueError:
                print('%s: Flow %s not found in foreground %s' % (j['entityId'], j['flow'], fg.ref))
                flow = LcFlow(j['flow'], Name=j['tags']['Name'], Compartment=['Intermediate Flows', 'Fragments'],
                              referenceQuantity=fg._catalog.get_canonical('mass'))
                fg.add(flow)
        frag = cls(j['entityId'], flow, j['direction'], origin=fg.ref, parent=parent,
                   exchange_value=j['exchangeValues'].pop('0'),
                   private=j['isPrivate'],
                   balance_flow=j['isBalanceFlow'])
        if j['externalId'] != j['entityId']:
            frag.external_ref = j['externalId']
        frag._exchange_values[1] = j['exchangeValues'].pop('1')
        for i, v in j['exchangeValues'].items():
            if i.find('____') >= 0:
                i = tuple(i.split('____'))
            frag._exchange_values[i] = v
            # frag.set_exchange_value(i, v)
        for tag, val in j['tags'].items():
            frag[tag] = val  # just a fragtag group of values
        return frag

    def finish_json_load(self, fg, j):
        # self.reference_entity = catalog[0][j['parent']]
        for k, v in j['terminations'].items():
            if k == 'default' or k == 'null':
                self.term_from_json(fg, None, v)
            else:
                self.term_from_json(fg, k, v)

    @classmethod
    def from_exchange(cls, parent, exchange):
        """
        This method creates a child flow, positioning the parent node as the 'process' component of the exchange
        and using the exchange's 'flow' and 'direction' components to define the child flow.  If the exchange
        also includes a 'termination', then that is used to automatically terminate the child flow.
        :param parent:
        :param exchange:
        :return:
        """
        frag = cls(uuid.uuid4(), exchange.flow, exchange.direction, parent=parent,
                   exchange_value=exchange.value, Name=exchange.flow['Name'])

        if exchange.termination is not None:
            frag.terminate(CatalogRef(exchange.flow.origin, exchange.termination, entity_type='process'),
                           term_flow=exchange.flow)
        return frag

    def __init__(self, the_uuid, flow, direction, parent=None,
                 exchange_value=1.0,
                 units=None,
                 private=False,
                 balance_flow=False,
                 background=False,
                 termination=None,
                 term_flow=None,
                 external_ref=None,
                 observe=None,
                 descend=None,
                 **kwargs):
        """
        Required params:
        :param the_uuid: use .new(Name, ...) for a random UUID
        :param flow: an LcFlow (catalog ref doesn't cut it)
        :param direction:
        :param parent:
        :param exchange_value: auto-set- cached; can only be set once
        :param units: to convert exchange value, if supplied
        :param private: forces aggregation of subfragments (not currently used)
        :param balance_flow: if true, exch val is always ignored and calculated based on parent
        :param background: [[DEPRECATED / to be ignored]] if true, ends traversal
        :param termination: specify the fragment's default termination
        :param term_flow: specify the term_flow (ignored if termination is None)
        :param observe: [None] if True, assign observed_ev to match cached_ev
        :param descend: [None] if a termination is provided, whether to set its descend flag
        :param kwargs:
        """

        self.__dbg_threshold = -1  # higher number is more verbose
        self._exchange_values = _new_evs()
        self._balance_child = None
        self._private = private
        # self._background = background
        self._is_balance = False
        self._child_flows = list()

        super(LcFragment, self).__init__('fragment', external_ref, entity_uuid=the_uuid, **kwargs)
        if self._external_ref == self._uuid:
            self._external_ref = None  # reset this

        if parent is not None:
            self.set_parent(parent)

        # # this is a problem when using a flow that is defined by a DelayedQuery that hasn't been resolved yet
        # assert flow.entity_type == 'flow', '%s %s' % (flow.entity_type, flow.link)
        # # this is a problem when referencing a flow with unknown origin / no query
        # assert flow.reference_entity.entity_type == 'quantity'
        self.flow = flow
        self.direction = check_direction(direction)  # w.r.t. parent

        self._terminations = dict()
        self._terminations[None] = FlowTermination.null(self)

        # set termination and StageName
        if termination is None:
            if 'StageName' not in self._d:
                self._d['StageName'] = ''
        else:
            self._terminations[None] = FlowTermination(self, termination, term_flow=term_flow, descend=descend)
            if 'StageName' not in self._d:
                stagename = ''
                try:
                    stagename = termination['Name']
                except (TypeError, KeyError):
                    try:
                        stagename = termination.name
                    except AttributeError:
                        pass
                finally:
                    self._d['StageName'] = stagename
        # set EV last
        if balance_flow:
            try:
                self.set_balance_flow()
            except RefQuantityRequired:
                # nothing much to do but notify
                logging.warning('%s: Balance flow missing reference entity' % self.link)
        else:
            if exchange_value is not None:
                self.set_exchange_value(0, exchange_value, units=units)
            if observe:
                self.observed_ev = self.cached_ev

    def __hash__(self):
        """
        Fragments must use UUID as hash because the link is mutable
        :return:
        """
        if self._origin is None:
            raise AttributeError('Origin not set!')
        return hash('%s/%s' % (self.origin, self.uuid))

    def resolve(self):
        if hasattr(self.flow, 'resolved') and not self.flow.resolved:
            self.dbg_print('attempting to resolve %s' % self.flow.link)
            self.flow = self.flow.resolve()
        if self.flow.reference_entity is None:
            self.dbg_print('failed to resolve %s' % self.flow.link)

    @property
    def external_ref(self):
        if self._external_ref is None:
            return self._uuid
        return self._external_ref

    @external_ref.setter
    def external_ref(self, ref):
        """
        Specify how the entity is referred to in the source dataset. If this is unset, the UUID is assumed
        to be used externally.
        :param ref:
        :return:
        """
        if self._external_ref is None:
            if ref != self.uuid:  # don't bother setting if it's the same as the UUID
                self._external_ref = ref
        else:
            raise PropertyExists('External Ref already set to %s' % self._external_ref)

    def de_name(self):
        """
        Remove a fragment's name
        :return:
        """
        self._external_ref = None

    def set_debug_threshold(self, level):
        self.__dbg_threshold = level

    def dbg_print(self, qwer, level=1):
        if level < self.__dbg_threshold:
            print('%.3s %s' % (self.uuid, qwer))

    def reference(self, flow=None):
        """
        For process interoperability
        :return:
        """
        rx = RxRef(self, self.flow, comp_dir(self.direction), self.get('Comment', None), value=self.observed_ev)
        if flow is not None:
            if not rx.flow.match(flow):
                raise ValueError('%.5s: Supplied flow %s does not match fragment' % (self.uuid, flow))
        return rx

    def make_ref(self, query):
        """
        We do NOT want to make local fragments into refs-- but we DO want to make remote fragments into refs. so
        we simply neuter the workhorse function here.
        :param query:
        :return:
        """
        '''
        ref = super(LcFragment, self).make_ref(query)
        ref.set_config(self.flow.make_ref(query.cascade(self.flow.origin)), self.direction)
        return ref
        '''
        return self

    def top(self):
        if self.reference_entity is None:
            return self
        return self.reference_entity.top()

    def set_parent(self, parent):
        if self.reference_entity is not None:
            self.unset_parent()
        if self.origin != parent.origin:
            if self.origin is None:
                self.origin = parent.origin
            else:
                raise AttributeError('Origin mismatch: parent (%s) vs child (%s)' % (parent.origin, self.origin))
        self._set_reference(parent)
        parent.add_child(self)

    def unset_parent(self):
        if self.is_balance:
            self.unset_balance_flow()
        self.reference_entity.remove_child(self)
        self._set_reference(None)

    '''
    def has_child(self, flow, direction, termination=None):
        for c in self.child_flows:
            if c.flow == flow and c.direction == direction:
                term = c.term
                if term.is_null:
                    if termination is None:
                        return True
                else:
                    if term.is_process and term.term_node.external_ref == termination:
                        return True
                    # this looks a bit ridiculous but I think it's right
                    if term.is_frag and term.term_node.term.term_node.external_ref == termination:
                        return True
        return False
    '''

    def add_child(self, child):
        """
        This should only be called from the child's set_parent function
        :param child:
        :return:
        """
        # if child.reference_entity is not self:
        #     raise InvalidParentChild('Fragment should list parent as reference entity')
        if child not in self._child_flows:
            self._child_flows.append(child)
        if self.term.is_null:  # formerly to_foreground()
            self.terminate(self)
        for term in self._terminations.values():
            term.clear_score_cache()

    def remove_child(self, child):
        """
        This should only be called from the child's unset_parent function
        :param child:
        :return:
        """
        if child.reference_entity is not self:
            raise InvalidParentChild('Fragment is not a child')
        self._child_flows.remove(child)
        if len(self._child_flows) == 0:
            if self.term.term_node is self:
                self.clear_termination()
        for term in self._terminations.values():
            term.clear_score_cache()

    @property
    def child_flows(self):
        """
        Yields the child flows in the order they were created.

        :return:
        """
        for k in self._child_flows:  # sorted(self._child_flows, key=lambda x: x.uuid):
            yield k

    @property
    def child(self):
        """
        little convenience function to avoid having to write next(frag.child_flows)
        :return:
        """
        return self._child_flows[0]

    def children_with_flow(self, flow, direction=None, termination=None, recurse=False):
        for k in self._child_flows:
            if k.flow == flow:
                if direction is not None:
                    if k.direction != direction:
                        continue
                if termination is not None:
                    if k.term.term_node != termination:
                        continue
                yield k
            if recurse:  # depth-first
                for z in k.children_with_flow(flow, direction, recurse=recurse):
                    yield z

    @property
    def parent(self):
        return self.reference_entity

    @property
    def is_reference(self):
        return self.reference_entity is None

    @property
    def term(self):
        return self._terminations[None]

    @property
    def name(self):
        if self._external_ref is None:
            return self['Name']
        return self.external_ref

    @property
    def dirn(self):
        return {
            'Input': '-<-',
            'Output': '=>='
        }[self.direction]

    def _serialize_evs(self):
        evs = dict()
        for k, v in self._exchange_values.items():
            if k is None:
                evs["0"] = v
            elif isinstance(k, int):
                evs[str(k)] = v
            else:
                evs[k] = v
        return evs

    def _serialize_terms(self, save_unit_scores=False):
        terms = dict()
        for k, v in self._terminations.items():
            if k is None:
                terms['default'] = v.serialize(save_unit_scores=save_unit_scores)
            else:
                terms[k] = v.serialize()
        return terms

    def serialize(self, save_unit_scores=False, domesticate=True, **kwargs):
        j = super(LcFragment, self).serialize(domesticate=True, **kwargs)  # once you save a fragment, it's yours

        f_e_r = {
            'origin': self.flow.origin,
            'externalId': self.flow.external_ref,
        }
        j.update({
            'entityId': self.uuid,
            'flow': f_e_r,
            'direction': self.direction,
            'isPrivate': self._private,
            'isBackground': self.is_background,  # descriptive only
            'isBalanceFlow': self.is_balance,
            'exchangeValues': self._serialize_evs(),
            'terminations': self._serialize_terms(save_unit_scores=save_unit_scores),
            'tags': self._d
        })
        for k in self._d.keys():
            j.pop(k, None)  # we put these together in tags
        return j

    @property
    def unit(self):
        """
        used for formatting the fragment in display
        :return:
        """
        if self.reference_entity is None:
            return '%4g %s' % (self.cached_ev, self.flow.unit)
        return self.term.unit

    def __str__(self):
        if self.reference_entity is None:
            if self.is_background:
                re = '(B) ref'
            else:
                re = ' ** ref'
        else:
            re = self.reference_entity.uuid[:7]
        if self.external_ref == self.uuid:
            extname = ''
        else:
            extname = '{%s}' % self.external_ref

        return '(%s) %s %.5s %s %s  [%s] %s %s' % (re, self.dirn, self.uuid, self.dirn, self.term,
                                                   self.unit, self['Name'], extname)

    def show(self):
        print('%s' % self)
        super(LcFragment, self).show()
        evs = list(self._exchange_values.keys())
        evs.remove(0)
        evs.remove(1)
        print('Exchange values: ')
        print('%20.20s: %g' % ('Cached', self.cached_ev))
        print('%20.20s: %g' % ('Observed', self.observed_ev))
        for k in evs:
            print('%20.20s: %g' % (k, self.exchange_value(k)))
        if self.is_balance:
            print('\nBalance flow: True (%s)' % self.flow.reference_entity)
        else:
            print('\nBalance flow: False')
        print('Terminations: ')
        print('%20s  %s' % ('Scenario', 'Termination'))
        for k, v in self._terminations.items():
            if v.term_node is self:
                print('%20.20s: %s Foreground' % (k, v))
            else:
                if v.descend:
                    desc = '     '
                else:
                    desc = '(agg)'
                print('%20.20s: %s %s %s' % (k, v, desc, v.term_node))

    def show_tree(self, scenario=None, observed=False, prefix=''):
        """
        THIS WHOLE FUNCTION is deprecated-- need to use live traversal results and not recurse over static data
        :param scenario:
        :param observed:
        :param prefix:
        :return:
        """
        children = [c for c in self.child_flows]
        term = self.termination(scenario)
        if len(children) > 0 and term.is_null:
            raise InvalidParentChild('null-terminated fragment %.7s has children' % self.uuid)

        delim = '()'
        if self.observed_ev != 0.0:
            delim = '[]'
        if not(observed and self.observed_ev == 0.0):
            # when doing the observed mode, don't print zero results
            print('   %s%s%s %.5s %s%s%7.3g %s%s %s' % (prefix, self.dirn, term, self.uuid,
                                                        delim[0],
                                                        self._mod(scenario),
                                                        self.exchange_value(scenario, observed=observed) or 0.0,
                                                        self.flow.unit,
                                                        delim[1],
                                                        self['Name']))
        # print fragment reference
        latest_stage = ''
        if len(children) > 0:
            print('   %s [%s] %s' % (prefix, term.unit, self['Name']))
            prefix += '    | '
            for c in sorted(children, key=lambda x: (x['StageName'], not x.term.is_null,
                                                     x.term.is_bg or x.term.term_is_bg)):
                if not c.balance_flow:
                    if observed and c.exchange_value(scenario, observed=observed) == 0:
                        continue
                if c['StageName'] != latest_stage:
                    latest_stage = c['StageName']
                    print('   %s %5s Stage: %s' % (prefix, ' ', latest_stage))
                c.show_tree(scenario=scenario, observed=observed, prefix=prefix)
            prefix = prefix[:-3] + ' x '
            print('   %s' % prefix)

    @property
    def cached_ev(self):
        ev = self._exchange_values[0]  # or 1.0  ## Why??? we need to allow zero values
        if ev is None:
            ev = 0.0
        return ev

    @cached_ev.setter
    def cached_ev(self, value):
        if self.cached_ev != 1.0:
            raise CacheAlreadySet('Set Value: %g (new: %g)' % (self.cached_ev, value))
        if value == self.cached_ev:
            return
        self._exchange_values[0] = value

    def reset_cache(self):
        """
        this must be done explicitly
        :return:
        """
        self._exchange_values[0] = 1.0

    def scale_evs(self, factor):
        """
        needed when foregrounding terminations
        :param factor:
        :return:
        """
        for k, v in self._exchange_values.items():
            self._exchange_values[k] = v * factor

    def clear_evs(self):
        self._exchange_values = _new_evs()

    @property
    def observed_ev(self):
        """

        :return: the observed exchange value
        """
        ''' # this is bogus
        ev = self._exchange_values[1]
        if self.reference_entity is None and ev == 0:  # now with named = observed, shouldn't every frag be observed?
            ev = self._exchange_values[0]
        return ev
        '''
        return self._exchange_values[1]

    @property
    def value(self):
        """
        for compatibility with Exchanges
        :return:
        """
        return self.observed_ev

    def observable(self, scenario=None):
        return self._check_observability(scenario=scenario)

    def _check_observability(self, scenario=None):
        if self.reference_entity is None:
            return True
        elif self.is_balance:
            self.dbg_print('observability: value set by balance.')
            return False
        elif self.reference_entity.termination(scenario).is_subfrag:
            self.dbg_print('observability: value set during traversal')
            return False
        else:
            return True

    @observed_ev.setter
    def observed_ev(self, value):
        if self._check_observability(None):
            self._exchange_values[1] = value

    def _observe(self, scenario=None, value=None, units=None):
        """
        no reason to have an interactive observe engine in an entity definition
        :param scenario:
        :return:
        """
        if value is None:
            '''
            if scenario is None:
                prompt = 'Observed value'
            else:
                prompt = 'Scenario value'

            print('%s' % self)
            print(' Cached EV: %6.4g\n Observed EV: %6.4g [%s]' % (self.cached_ev, self.observed_ev, self.flow.unit))
            if scenario is None:
                string_ev = '%10g' % self.observed_ev
            else:
                string_ev = '%10g' % self.exchange_value(scenario)
                print(' Scenario EV: %s [%s]' % (string_ev,
                                                 self.flow.unit))

            val = ifinput('%s ("=" to use cached): ' % prompt, string_ev)

            if val != string_ev:
                if val == '=':
                    value = self.cached_ev
                else:
                    value = parse_math(val)
            '''
            raise ValueError('%.5s: Either supply an observed value or accept_all=True' % self.uuid)

        if scenario is None:
            self.set_exchange_value(1, value, units=units)
        else:
            self.set_exchange_value(scenario, value, units=units)

    def _auto_observe(self, scenario=None):
        if scenario is None:
            self.observed_ev = self.cached_ev
        else:
            self.set_exchange_value(scenario, self.cached_ev)

    def observe(self, value=None, units=None, scenario=None, accept_all=False, recurse=True, _traverse=True):
        """
        Report the fragment's observed exchange value and assign it a name
        if fragment is a balance flow or if fragment is a child of a subfragment (for the specified scenario), then
         the ev is set during traversal and may not be observed.

        :param value: the observed exchange value (if omitted, and accept_all is False, observe interactively)
        :param units: optional unit for observed value
        :param scenario:
        :param accept_all: whether to automatically apply the cached EV to the observation
        :param recurse: whether to observe child fragments. automatically False if a value is supplied
        :param _traverse: internal param. used to only traverse the top-most fragment.
        :return:
        """
        if self._check_observability(scenario=scenario):
            if accept_all and (value is None):  # ignore
                self._auto_observe(scenario=scenario)
            else:
                self._observe(scenario=scenario, value=value, units=units)

        if recurse and (value is None):
            for c in self.child_flows:
                c.observe(scenario=scenario, accept_all=accept_all, recurse=True, _traverse=False)

        if _traverse:
            self.traverse(scenario, observed=True)

    @property
    def is_background(self):
        return len(self._child_flows) == 0  # self._background

    def scenarios(self, recurse=True):
        """
        Generate a list of scenarios known to the fragment.

        :param recurse: [True] By default, traverse all child flows, including subfragments. If False, only report
         the present fragment
         filter out internal scenarios (start with '__') from recursive queries but not from the top level
        :return:
        """
        scenarios = set(list(self._exchange_values.keys()))
        for scen, term in self._terminations.items():
            scenarios.add(scen)
            if recurse and term.is_subfrag:
                for k in term.term_node.scenarios(recurse=True):
                    if str(k).startswith('__'):
                        continue
                    scenarios.add(k)

        if recurse:
            for c in self.child_flows:
                for k in c.scenarios(recurse=True):
                    scenarios.add(k)

        scenarios -= {0, 1, None}
        for k in sorted(scenarios):
            yield k

    def clear_scenarios(self, terminations=True):
        sc = [k for k in self._exchange_values.keys() if k not in (0, 1)]
        for s in sc:
            self._exchange_values.pop(s)
        if terminations:
            sc = [k for k in self._terminations.keys() if k is not None]
            for s in sc:
                self._terminations.pop(s)

    def _match_scenario_ev(self, scenario):
        """
        If *any* scenario is specified, we return either a match or the observed ev.

        Special case: if the matched scenario begins with 'norm' AND the current fragment is a reference fragment,
        then interpret the scenario as a normalization factor and REMOVE it from subsequent traversals to prevent
        double-norming. Not totally convinced this is a good idea, but it is useful for some select circumstances.

        Plus, double-norming (in recursive descent) is objectively wrong-- there is no circumstance when it is right.
        :param scenario:
        :return:
        """
        if scenario is None:
            return None
        if isinstance(scenario, set):
            match = [scen for scen in filter(None, scenario) if scen in self._exchange_values.keys()]
            if len(match) == 0:
                return 1
            elif len(match) > 1:
                raise ScenarioConflict('fragment: %s\nexchange value matches: %s' % (self, match))
            m = match[0]
            if str(m).startswith('norm') and self.parent is None:
                print('Applying and removing one-time normalization scenario %s' % m)
                # self.dbg_print('Applying and removing one-time normalization scenario %s' % m, level=0)
                scenario.remove(m)
            return m
        if scenario in self._exchange_values.keys():
            return scenario
        return 1

    def _match_scenario_term(self, scenario):
        if scenario == 0 or scenario == '0' or scenario is None:
            return None
        if isinstance(scenario, set):
            match = [scen for scen in filter(None, scenario) if scen in self._terminations.keys()]
            if len(match) == 0:
                return None
            elif len(match) > 1:
                raise ScenarioConflict('fragment: %s\ntermination matches: %s' % (self, match))
            return match[0]
        if scenario in self._terminations.keys():
            return scenario
        return None

    def exchange_value(self, scenario=None, observed=False):
        """

        :param scenario: None, a string, or a tuple of strings. If tuple, raises error if more than one match.
        :param observed:
        :return:
        """
        if scenario is None:
            if observed:
                ev = self.observed_ev
            else:
                ev = self.cached_ev
        else:
            if isinstance(scenario, list) or isinstance(scenario, tuple):
                scenario = set(scenario)
            match = self._match_scenario_ev(scenario)
            if match is None:
                ev = self.observed_ev
            else:
                ev = self._exchange_values[match]
        if ev is None:
            return 0.0
        return ev

    def parameters(self):
        p = {'origin': self.origin, 'parameter': self.external_ref, 'flow_unit': self.flow.unit}
        if self.observable():
            p['default'] = self.observed_ev
        for k, v in self._exchange_values.items():
            if k in (0, 1):  # this obviously needs to be redone
                continue
            if self.observable(k):
                p[k] = v
        return p

    def _mod(self, scenario):
        """
        Returns a visual indicator that reports whether the fragment's exchange_value or termination are affected
        under a given scenario.
        :param scenario:
        :return: ' ' no effect
                 '=' balance flow
                 '+' termination affected under scenario
                 '*' exchange value affected under scenario
                 '%' both termination and exchange value affected under scenario
        """
        if self.is_balance:
            return '='
        match_e = self._match_scenario_ev(scenario)
        match_t = self._match_scenario_term(scenario)
        if match_e is None or self.exchange_value(match_e) == self.cached_ev:
            if match_t is None:
                return ' '  # no scenario
            return '+'  # term scenario
        if match_t is None:
            return '*'  # ev scenario
        return '%'  # both scenario

    def set_exchange_value(self, scenario, value, units=None):
        """
        The exchange value may not be set on a reference fragment UNLESS the termination for the named scenario is
        to foreground- in which case the term's inbound_ev is set instead. (and a term is created if none exists)
        :param scenario:
        :param value: if 'None', un-set any observation for the given scenario
        :param units: unit string, if value is to be converted from non-reference units
        :return:
        """
        if value is None:
            if scenario not in {1, '1', 0, '0', None}:
                v = self._exchange_values.pop(scenario, None)
                if v:
                    self.dbg_print('Removed observation %g for scenario %s' % (v, scenario), level=0)
            return
        if not self._check_observability(scenario=scenario):
            raise DependentFragment('Fragment exchange value set during traversal')
        if isinstance(scenario, tuple) or isinstance(scenario, set):
            raise ScenarioConflict('Set EV must specify single scenario')

        value = float(value)

        if units is not None and len(units) > 0:
            try:
                value *= self.flow.reference_entity.convert(units)
            except ConversionError:
                print('##!! Flow conversion error setting exchange value: %5.5s: %s (%s)' % (self.uuid,
                                                                                             self.flow.reference_entity,
                                                                                             units))
                value = 0.0

        if scenario == 0 or scenario == '0' or scenario == 'cached' or scenario is None:
            self._exchange_values[0] = value
        elif scenario == 1 or scenario == '1' or scenario == 'observed':
            self._exchange_values[1] = value
        else:
            self._exchange_values[scenario] = value

    @property
    def conserved(self):
        return bool(self.balance_magnitude)  # None or 0 -> False

    @property
    def balance_flow(self):
        return self._balance_child

    @property
    def is_balance(self):
        return self._is_balance

    def reverse_direction(self):
        """
        Changes the direction of a fragment to its complement, since the direction of the fragment flow is arbitrary.
        negates all stored exchange values, so as to have no effect on traversal computations.
        :return:
        """
        d = dict()
        for k, v in self._exchange_values.items():
            d[k] = -1 * v
        self.direction = comp_dir(self.direction)
        self._exchange_values = d

    def set_balance_flow(self):
        """
        A balance flow balances its own reference quantity.
        :return:
        """
        if self.reference_entity is None:
            raise InvalidParentChild('Reference flow cannot be a balance flow')
        if self.is_balance is False:
            self.reference_entity.set_conservation_child(self)
            self._is_balance = True

    def unset_balance_flow(self):
        if self.is_balance:
            self.reference_entity.unset_conservation_child()
            self._is_balance = False

    def set_conservation_child(self, child):
        if child.reference_entity != self:
            raise InvalidParentChild
        if self.is_conserved_parent and (child is not self._balance_child):
            print('%.5s conserving %s\nversus %s' % (self.uuid, self._balance_child, child))
            raise BalanceAlreadySet
        self._balance_child = child
        if self.balance_magnitude == 0:
            self.dbg_print('%.5s Notice: zero balance for conserved quantity %s' % (self.uuid,
                                                                                    self.conserved_quantity))
        self.dbg_print('setting balance from %.5s: %s' % (child.uuid, self._balance_child))

    @property
    def is_conserved_parent(self):
        return self._balance_child is not None

    def unset_conservation_child(self):
        self._balance_child = None

    @property
    def conserved_quantity(self):
        if self._balance_child is None:
            return None
        else:
            return self._balance_child.flow.reference_entity

    @property
    def balance_magnitude(self):
        """
        The cf of the current flow in balanced units (if a balance flow exists)
        :return:
        """
        if self._balance_child is None:
            return None
        else:
            if self._balance_child.flow.reference_entity is None:
                self._balance_child.resolve()
                logging.warning('tryna resolve %s' % self._balance_child.flow.link)
            f = self._balance_child.flow
            if f.reference_entity is None:
                self.dbg_print('Balance Flow %5.5s has no reference' % f.uuid, 0)
                raise RefQuantityRequired('Flow %s [%s] has no reference quantity' % (f.link, type(f)))
            try:
                cf = self._balance_child.flow.reference_entity.cf(self.flow)
                self.dbg_print('Found balance magnitude %g' % cf)
                # if cf == 0:  # debug goes here
                #     logging.error('%s' % zz)
                return cf
            except (QuantityRequired, MissingResource) as e:
                self.dbg_print('Encountered %s during cf' % e)
                return 0.0

    '''
    def balance(self, scenario=None, observed=False):
        """
        display a balance the inputs and outputs from a fragment termination.  This probably needs to be reimagined and
        rewritten
        :param scenario:
        :param observed:
        :return: a dict of quantities to balance magnitudes (positive = input to term node)
        """
        qs = defaultdict(float)
        if self.reference_entity is None:
            in_ex = self.exchange_value(scenario, observed=observed)
        else:
            in_ex = 1.0
        for cf in self.flow.reference_entity.profile(self.flow):
            if cf.value is not None:
                if self.direction == 'Input':  # output from term
                    qs[cf.quantity] -= cf.value * in_ex
                else:
                    qs[cf.quantity] += cf.value * in_ex
        for c in self.child_flows:
            for cf in c.flow.reference_entity.profile(c.flow):
                mag = c.exchange_value(scenario, observed=observed) * (cf.value or 0.0)
                if mag != 0:
                    if c.direction == 'Output':
                        qs[cf.quantity] -= mag
                    else:
                        qs[cf.quantity] += mag

        for k, v in qs.items():
            print('%10.4g %s' % (v, k))
        return qs

    def show_balance(self, quantity=None, scenario=None, observed=False):
        def _p_line(f, m, d):
            try:
                # will fail if m is None or non-number
                print(' %+10.4g  %6s  %.5s %s' % (m, d, f.uuid, f['Name']))
            finally:
                pass

        if quantity is None:
            quantity = self.flow.reference_entity

        print('%s' % quantity)
        mag = quantity.cf(self.flow).value
        if self.reference_entity is None:
            mag *= self.exchange_value(scenario, observed=observed)
        if self.direction == 'Input':
            mag *= -1

        net = mag

        _p_line(self, mag, comp_dir(self.direction))

        for c in sorted(self.child_flows, key=lambda x: x.direction):
            mag = c.exchange_value(scenario, observed=observed) * quantity.cf(c.flow).value
            if c.direction == 'Output':
                mag *= -1
            if mag is None or mag != 0:
                _p_line(c, mag, c.direction)
            net += mag

        print('----------\n %+10.4g net' % net)
    '''

    '''
    Terminations and related functions
    '''

    def terminate(self, term_node, scenario=None, direction=None, **kwargs):
        """
        specify the fragment's termination in a given scenario.
        :param term_node: The thing that terminates the flow
        :param scenario:
        :param direction: outmoded option (can still use _direction if kluge is needed)
        :param kwargs: term_flow, descend
        :return:
        """
        if scenario in ('cached', 'observed'):
            raise ValueError('scenario cannot use reserved name: %s' % scenario)

        if direction is not None:
            raise NotImplementedError('Flow Termination Directions are not supposed to be specified by hand')
        if isinstance(scenario, tuple) or isinstance(scenario, set):
            raise ScenarioConflict('Set termination must specify single scenario')
        if scenario is not None and scenario in self._terminations:
            if not self._terminations[scenario].is_null:
                raise CacheAlreadySet('Scenario termination already set. use clear_termination()')

        if scenario is None and term_node in self._terminations:
            # shortcut: assign an existing scenario to be default  ## DWR? this is surely nonstandard-
            # causes namespace conflicts btwn scenario and external_ref, but not really because the argument is
            # supposed to be a node. supplying txt as an argument is nonstandard- so it's clearly a hack. eh.
            self._terminations[None] = self._terminations[term_node]
            return self._terminations[term_node]

        termination = FlowTermination(self, term_node, **kwargs)
        self._terminations[scenario] = termination
        if scenario is None:
            if self['StageName'] == '' and not termination.is_null:
                if termination.is_frag:
                    self['StageName'] = termination.term_node['StageName']
                elif termination.is_context:
                    self['StageName'] = termination.term_node.name
                else:
                    try:
                        self['StageName'] = termination.term_node['Name']
                    except (KeyError, TypeError, IndexError):
                        print('%.5s StageName failed %s' % (self.uuid, termination.term_node))
                        self['StageName'] = termination.term_node.name
        return termination

    def clear_termination(self, scenario=None):
        self._terminations[scenario] = FlowTermination.null(self)

    '''
    def to_foreground(self, scenario=None):
        """
        make the fragment a foreground node. This is done by setting the termination to self (only if the term is
        currently null).  A foreground node
        may not be a background node (obv.)  Note: a foreground node will not act as an emission during traversal and
        fragment LCIA (it must be terminated to a context), but it can have observed unit scores.
        However, it will not show up as a cutoff.  foreground() fragments can conceal flows.
        :param scenario:
        :return:
        """
        # self.unset_background()
        if self.termination(scenario).is_null:
            self.terminate(self, scenario=scenario)
    '''

    '''
    def unset_background(self):
        if self._background is True:
            for term in self._terminations.values():
                term.clear_score_cache()
        self._background = False

    def set_background(self):
        for scenario, term in self._terminations.items():
            if not term.is_null:
                if term.term_node.entity_type == 'fragment':
                    raise ScenarioConflict('Cannot bg: Terminated to fragment in Scenario %s' % scenario)
        if self._background is False:
            for term in self._terminations.values():
                term.clear_score_cache()
        self._background = True
    '''

    def term_from_json(self, fg, scenario, j):
        if isinstance(scenario, tuple):
            raise ScenarioConflict('Set termination must specify single scenario')
        self._terminations[scenario] = FlowTermination.from_json(self, fg, scenario, j)

    def termination(self, scenario=None):
        match = self._match_scenario_term(scenario)
        if match in self._terminations.keys():
            return self._terminations[match]
        # if None in self._terminations.keys():  # this should be superfluous, as match will be None
        #     return self._terminations[None]
        raise ScenarioConflict('This should never happen - No match found for %s' % scenario)

    def terminations(self):
        return self._terminations.items()

    '''
    def set_child_exchanges(self, scenario=None, reset_cache=False):
        """
        This is really client code, doesn't use any private capabilities
        Set exchange values of child flows based on inventory data for the given scenario.  The termination must be
         a foreground process.

        In order for this function to work, flows in the node's exchanges have to have the SAME external_ref as the
        child flows, though origins can differ.  There is no other way for the exchanges to be set automatically from
        the inventory.  Requiring that the flows have the same name, CAS number, compartment, etc. is too fragile /
        arbitrary.  The external refs must match.

        This works out okay for databases that use a consistent set of flows internally -- ILCD, thinkstep, and
        ecoinvent all seem to have that characteristic but ask me again in a year.

        To automatically set child exchanges for different scenarios that use processes from different databases,
        encapsulate each term node inside a sub-fragment, and then specify different subfragment terminations for the
        different scenarios.  Then, map each input / output in the sub-fragment to the correct foreground flow using
        a conserving child flow.

        In that case, the exchange values will be set during traversal, and each sub-fragment's internal exchange
        values can be set automatically using set_child_exchanges.

        :param scenario: [None] for the default scenario, set observed ev
        :param reset_cache: [False] if True, for the default scenario set cached ev
        :return:
        """
        term = self.termination(scenario)
        if not term.term_node.entity_type == 'process':
            raise DependentFragment('Child flows are set during traversal')

        if scenario is None and self.reference_entity is None:
            # this counts as observing the reference flow
            if self.observed_ev == 0:
                self.observed_ev = self.cached_ev

        children = defaultdict(list)  # need to allow for differently-terminated child flows -- distinguish by term.id

        for k in self.child_flows:
            key = (k.flow.external_ref, k.direction)
            children[key].append(k)
        if len(children) == 0:
            return

        for x in term.term_node.inventory(ref_flow=term.term_flow, direction=term.direction):
            if x.value is None:
                self.dbg_print('skipping None-valued exchange: %s' % x)
                continue

            key = (x.flow.external_ref, x.direction)
            if key in children:
                try:
                    if len(children[key]) > 1:
                        child = next(c for c in children[key] if c.termination(scenario).id == x.termination)
                    else:
                        child = next(c for c in children[key])
                except StopIteration:
                    continue

                self.dbg_print('setting %s [%10.3g]' % (child, x.value))
                if scenario is None:
                    if reset_cache:
                        child.reset_cache()
                        child.cached_ev = x.value
                    else:
                        child.observed_ev = x.value
                else:
                    child.set_exchange_value(scenario, x.value)

    def _node_weight(self, magnitude, scenario):
        term = self.termination(scenario)
        '.'.'.
        if self.reference_entity is None and term.is_fg:
            return magnitude / self.exchange_value(scenario, observed=observed)
        '.'.'.
        if term is None or term.is_null:
            return magnitude

        else:
            return magnitude * term.node_weight_multiplier

    def _cache_balance_ev(self, _balance, scenario, observed):
        """
        TL;DR: need to stop caching balance flows, including subfragment traversal results, and move towards using
        live traversals (i.e. lists of FragmentFlows) for EVERYTHING from show_tree() to reports. For the time
        being, the cached values will report whatever was the LAST cached value-- so if a fragment is un-traversed
        in a certain scenario, particularly if the scenario affects a sub-fragment and not the local fragment,
        it will retain a false cache value.

        Still not totally clear on live traversals-- but maybe we simply SHOULDN'T cache if match is None.  But really
        we should move away from querying cached exchange values.  And we definitely shouldn't serialize exchange values
        for un-observable fragments.

        I think we should actually raise an exception when exchange value is queried for non-None scenario on a non-
        observable flow, requiring client code to use the FragmentFlow.

        BIG FAT BUG: evs can be modified by scenarios not defined locally in the current fragment.  Ergo, checking to
        see if the fragment's ev dict has the given scenario is not sufficient- we should not be setting the
        'observed ev' when any scenario is in effect. the whole scenario tuple needs to be used. this is a cheap dict,
        after all.  For balancing + fragment child flows only.  so this needs to be thought through somewhat.
        ans- no it doesn't! if they are balance / fffc flows, then their ev is never used! the ev dict is only for
        recordkeeping!  except set_exchange_value is limited to one scenario. so- don't use it.
        :param _balance:
        :param scenario:
        :return:
        """
        if scenario is None:
            if observed:
                self._exchange_values[1] = _balance
            else:
                self._exchange_values[0] = _balance
        else:
            match = self._match_scenario_ev(scenario)
            if match is not None:
                self._exchange_values[match] = _balance
    '''
    def nodes(self, scenario=None, descend=True, group='StageName'):
        """
        Report proximal terminal nodes for the fragment (recurse until a nondescend is reached)
        :param scenario: [None]
        :param descend: [True] if False, yield subfragments as nodes
        :return: generator of terminal nodes
        """
        term = self.termination(scenario)
        ev = self.exchange_value(scenario)
        gp = self.get(group, '')
        if term.is_process or term.is_context or term.is_unresolved:
            yield FragmentBranch(self, term, gp, scenario=scenario, magnitude=ev, is_cutoff=False)
        elif term.is_subfrag:
            if term.descend and descend:
                for n in term.term_node.nodes(scenario, descend=descend):
                    yield n
            else:
                yield FragmentBranch(self, term, gp, scenario=scenario, magnitude=ev, is_cutoff=False)
            # foreground, null: do nothing
        for c in self.child_flows:
            for n in c.nodes(scenario, descend=descend):
                yield n

    def tree(self):
        """
        This is real simple- just a recursive enumeration of child flows, depth first

        :return:
        """
        yield self
        for c in sorted(self.child_flows, key=lambda x: (x['StageName'], not x.term.is_null, x.term.is_bg)):
            for b in c.tree():
                yield b

    def fragment_lcia(self, quantity_ref, scenario=None, observed=True, mode=None, group_by=None, **kwargs):
        """
        Fragments don't have access to a qdb, so this piggybacks on the quantity_ref.
        note: refresh is no longer supported during traversal
        :param quantity_ref:
        :param scenario:
        :param observed: [True] whether to limit the computation to observed flows
        :param mode: None, 'detailed' (NOP when local), 'flat', 'stage', 'anchor'
        :param group_by: fragment property to group by for stage_lcia
        :param kwargs: ultimately passed down to LCIA computation routine
        :return:
        """
        fragmentflows = self.traverse(scenario=scenario, observed=observed)
        res = frag_flow_lcia(fragmentflows, quantity_ref, scenario=scenario, **kwargs)
        if mode == 'flat':
            return res.flatten()
        elif mode == 'stage':
            if group_by:
                return res.aggregate(key=lambda x: x.fragment.get(group_by, 'Others'))
            return res.aggregate()
        elif mode == 'anchor':
            return res.terminal_nodes()
        else:
            return res

    def activity(self, scenario=None, observed=True):
        """
        Reports a list of node activities of direct descendants only
        :param scenario:
        :param observed:
        :return:
        """
        top = self.top()

        ffs = top.traverse(scenario=scenario, observed=observed)

        act = [k for k in ffs if k.fragment.top() is top]
        return act

    def inventory(self, scenario=None, scale=1.0, observed=False):
        raise FragmentInventoryDeprecated('"inventory" is an exchange method. Use "cutoffs" instead.')

    def cutoffs(self, scenario=None, scale=1.0, observed=False):
        """
        Converts unit inventory into a set of exchanges for easy display
        :param scenario:
        :param scale:
        :param observed:
        :return:
        """
        ios, _ = self.unit_flows(scenario=scenario, observed=observed)
        return ios_exchanges(ios, ref=self)

    def exchanges(self, scenario=None):
        """
        Generator for query compatibility with processes
        :param scenario:
        :return:
        """
        for x in self.child_flows:
            yield x
        for x in self.termination(scenario).unobserved_exchanges():
            yield x

    def unit_inventory(self, scenario=None, observed=False, frags_seen=None):
        raise FragmentInventoryDeprecated('"inventory" is an exchange method. "unit_flows" replaces unit_inventory.')

    def unit_flows(self, scenario=None, observed=False, frags_seen=None):
        """
        Traverses the fragment containing self, and returns a set of FragmentFlows indicating the net input/output
         with respect to a *unit node weight of the reference fragment*.

        Within the set of fragment flows:
         * all will have null terminations
         * every flow appears with only one direction
         * the fragment's reference flow will appear with a direction relative to the fragment.

        Created to encapsulate a traversal problem.

        :param scenario:
        :param observed:
        :param frags_seen: to prevent recursive loops
        :return: list of io flows,
        """
        top = self.top()

        if frags_seen:
            frags_seen = set(frags_seen)  # reset to allow reentry

        ffs = top.traverse(scenario, observed=observed, frags_seen=frags_seen)

        ios, internal = group_ios(self, ffs)

        return ios, internal

    '''
    def cutoffs(self, scenario=None, observed=False, aggregate=True):
        """
        Return a comprehensive list of cut-offs from a traversal result. Include implicit cutoffs from background
         computations.
        :param scenario:
        :param observed:
        :param aggregate: [True] if True, group cutoffs from different nodes together by flow.  If False, report each
        cutoff from a distinct node distinctly.
        :return:
        """
        ffs = self.traverse(scenario, observed=observed)
        cos = []
        for ff in ffs:
            if ff.term.is_null:
                cos.append(ff)
            else:
                if ff.fragment.is_background:
                    #extend ff with background node cut-off flows.
                    #Need to think about this for a minute because self.is_background could be terminated to either
                    #a process or a fragment.  If it's a process, then for fragment LCIA we will be computing bg_lcia,
                    #so assuming a CatalogRef with background access.
                    
                    ref = ff.term.term_node
                    cos.extend([FragmentFlow.cutoff(ff.fragment, i.flow, i.direction, i.value * ff.node_weight)
                                for i in ref.lci(ref_flow=ff.term.term_flow.external_ref)
                                if i.type in ('cutoff', 'context')])
                elif ff.term.is_process:
                    # TODO: add in cutoffs from unobserved exchanges
                    pass

        if aggregate:
            cos, _ = group_ios(self, cos, include_ref_flow=False)

        return sorted([ExchangeValue(f.fragment, f.fragment.flow, f.fragment.direction, value=f.magnitude)
                       for f in cos], key=lambda x: (x.direction == 'Input', x.flow.context,
                                                     x.flow['Name'], x.value), reverse=True)
    '''

    def traverse(self, scenario=None, observed=False, frags_seen=None):
        if isinstance(scenario, set):
            scenarios = set(scenario)
        elif isinstance(scenario, tuple) or isinstance(scenario, list):
            scenarios = set(scenario)
        elif scenario is None:
            scenarios = set()
        else:
            scenarios = {scenario}
        if len(scenarios) == 0:
            if observed:
                scenarios = {1}
            else:
                scenarios = None
        ffs, _ = self._traverse_node(1.0, scenarios, frags_seen=frags_seen)
        return ffs

    def _traverse_fg_node(self, ff, scenarios, frags_seen):
        """
        Handle foreground nodes and processes--> these can be quantity-conserving, but except for
        balancing flows the flow magnitudes are determined at the time of construction (or scenario specification).

        Balancing flow exchange values are always determined with respect to a unit activity of the terminal node.

        This is messy so it deserves some notes.
        the fragment's exchange value specifies the scaling factor for the terminal node, EXCEPT if the
        fragment is a reference fragment (no parent) AND the terminal node is in the foreground.  In this case
        ONLY, the exchange value functions as an inbound exchange value, and the node weight should be 1.0.
        The prototypical example of this is a fragment for cultivation of 1 ha of corn, with the reference
        flow being (e.g.) 9300 kg of corn.  The reference fragment's exchange value is 9300 kg; the reference
         node's node weight is 1.0

        If the reference node has a balancing flow (say, we want to balance carbon content), then the stock
        value is the reference flow converted to the conservation quantity, e.g. 9300 * 0.43 = 3999 kg C, sign
        negative since it's going out (note that the reference fragment's direction is 'Input' in this case
        because direction is always interpreted relative to the parent).

        So then we traverse the child flows, let's say none of them have kg C characterization, and so our stock
        remains -3999 kg. WHen we hit the balancing fragment "atmospheric carbon in", we catch the FoundBalanceFlow
        and come back to it.

        When we come back, we re-negate the stock to +3999 and pass that as _balance to the balancing flow, which
        becomnes that flow's exchange value (again w.r.t. this node's unit node weight).

        IF the terminal node is a process, or if the node is an interior (non-reference) fragment, it's much easier.
        The stock is simply the process's inbound exchange value (with respect to a unit activity level), or if
        it's a foreground node then the stock is simply 1, and the node_weight already accounts for the exchange
        value and scales the balancing flow correctly.

        Were we to do this non-'live' or with some kind of stack, we could conceivably balance multiple quantities with
        additional fragment flows, but I'm pretty sure the same effect can be achieved by nesting conserving fragments.
        This also requires the modeler to specify a resolution order and cuts the need for elaborate error checking.

        :param ff: a FragmentFlow containing the foreground termination to recurse into
        :param scenarios:
        :param frags_seen:
        :return: a list of FragmentFlows in the order encountered, with input ff in position 0
        """
        term = ff.term
        node_weight = ff.node_weight
        ffs = [ff]
        stock_inflow = stock_outflow = 0.0  # keep track of total inflows to detect+quash near-zero (floating-point-tolerance) balances

        if term.is_fg or not term.valid:
            if self.reference_entity is None:
                # inbound exchange value w.r.t. term node's unit magnitude
                stock = self.exchange_value(scenarios)
            else:
                stock = 1.0  # balance measurement w.r.t. term node's unit magnitude
        else:
            stock = term.inbound_exchange_value  # balance measurement w.r.t. term node's unit magnitude
        bal_f = None
        if self.is_conserved_parent:
            stock *= self.balance_magnitude
            if self.direction == 'Input':  # convention: inputs to self are positive
                #  direction w.r.t. parent so self.direction == 'Input' is an input to parent = output from node
                stock *= -1
            self.dbg_print('%g inbound-balance' % stock, level=2)

        if stock > 0:
            stock_inflow += stock
        else:
            stock_outflow -= stock
        # TODO: Handle auto-consumption!  test child flows for flow identity and null term; cumulate; scale node_weight

        for f in self.child_flows:
            try:  # now that we are storing _balance_child we can just skip it instead of try-catch. but whatever.
                # traverse child, collecting conserved value if applicable
                child_ff, cons = f._traverse_node(node_weight, scenarios,
                                                  frags_seen=frags_seen, conserved_qty=self.conserved_quantity)
                if cons is None:
                    self.dbg_print('-- returned cons_value', level=3)
                else:
                    self.dbg_print('%g returned cons_value' % cons, level=2)
                    stock += cons
                    if cons > 0:
                        stock_inflow += cons
                    else:
                        stock_outflow -= cons
            except FoundBalanceFlow:
                self.dbg_print('%g bal magnitude on %.3s' % (stock, f.uuid), level=3)
                bal_f = f
                child_ff = []

            ffs.extend(child_ff)

        stock_max = max([abs(stock_inflow), abs(stock_outflow)])

        if bal_f is not None:
            # balance reports net inflows; positive value is more coming in than out
            # if balance flow is an input, its exchange must be the negative of the balance
            # if it is an output, its exchange must equal the balance
            if stock != 0 and (abs(stock) / stock_max < 1e-10):
                self.dbg_print('Quashing <1e-10 balance flow (%g vs %g)' % (stock, stock_inflow), level=1)
                stock = 0.0
            if bal_f.direction == 'Input':
                stock *= -1
                self.dbg_print('%.3s Input: negating balance value' % bal_f.uuid)
            else:
                self.dbg_print('%.3s Output: maintaining balance value' % bal_f.uuid)
            self.dbg_print('%g balance value passed to %.3s' % (stock, bal_f.uuid))
            bal_ff, _ = bal_f._traverse_node(node_weight, scenarios,
                                             frags_seen=set(frags_seen), conserved_qty=None, _balance=stock)
            ffs.extend(bal_ff)

        return ffs

    def _traverse_subfragment(self, ff, scenarios, frags_seen):
        """
        handle sub-fragments, including background flows--
        for sub-fragments, the flow magnitudes are determined at the time of traversal and must be pushed out to
         child flows
        for LOCAL background flows, the background ff should replace the current ff, maintaining self as fragment

        subfragment activity level is determined as follows:
         - if the subfragment is a background fragment, it MUST have a unity inbound_exchange_value; this is enforced:
           - for background processes, because FlowTermination._unobserved_exchanges() uses term_node.lci(ref_flow)
           - for background fragments, in the yet-to-be-implemented bg_lcia method  (part of the foreground interface??)
           in any case, the background term is swapped into the foreground node.

         - otherwise, the subfragment inventory is taken and grouped, and the matching flow is found, and the magnitude
           of the matching flow is used to normalize the downstream node weight.

         - then the inventory of the subfragment is used to apply exchange values to child fragments, and traversal
           recursion continues.


        :param ff: a FragmentFlow containing the non-fg subfragment termination to recurse into
        :param scenarios:
        :param frags_seen:
        :return:
        """
        '''
        '''

        term = ff.term
        if 0:  #  term.term_is_bg: this is deprecated
            pass
            '''
            if term.term_flow == term.term_node.flow:
                # collapse trivial bg terminations into the parent fragment flow
                bg_ff, _ = term.term_node._traverse_node(ff.node_weight, scenario, observed=observed)
                assert len(bg_ff) == 1, (self.uuid, term.term_node.external_ref)
                bg_ff[0].fragment = self
                return bg_ff
            '''

        # traverse the subfragment, match the driven flow, compute downstream node weight and normalized inventory
        try:
            ffs, unit_inv, downstream_nw = _do_subfragment_traversal(ff, scenarios, frags_seen)
        except ZeroInboundExchange:
            self.dbg_print('subfragment divide by zero', 1)
            return [ff]

        # next we traverse our own child flows, determining the exchange values from the normalized unit inventory
        for f in self.child_flows:
            self.dbg_print('Handling child flow %s' % f, 4)
            ev = 0.0
            try:
                m = next(j for j in unit_inv if j.fragment.flow == f.flow)
                if m.fragment.direction == f.direction:
                    self.dbg_print('  ev += %g' % m.magnitude, 4)
                    ev += m.magnitude
                else:
                    self.dbg_print('  ev -= %g' % m.magnitude, 4)
                    ev -= m.magnitude
                unit_inv.remove(m)
            except StopIteration:
                self.dbg_print('  no driving flows found')
                continue

            self.dbg_print('traversing with ev = %g' % ev, 4)
            child_ff, _ = f._traverse_node(downstream_nw, scenarios,
                                           frags_seen=frags_seen, _balance=ev)
            ffs.extend(child_ff)

        # remaining un-accounted io flows are getting appended, so do scale
        for x in unit_inv:
            x.scale(downstream_nw)
        ffs.extend(list(unit_inv))

        return ffs

    def _traverse_node(self, upstream_nw, scenarios,
                       frags_seen=None, conserved_qty=None, _balance=None):

        """
        If the node has a non-null termination, use that; follow child flows.

        If the node's termination is null- then look for matching background fragments. If one is found, adopt its
        termination, and return.

        else: assume it is a null foreground node; follow child flows

        :param upstream_nw: upstream node weight
        :param scenarios: set of scenario values
        #:param observed: whether to use observed or cached evs (overridden by scenario specification)
        :param frags_seen: carried along to catch recursion loops
        :param conserved_qty: in case the parent node is a conservation node
        :param _balance: used when flow magnitude is determined during traversal, i.e. for balance flows and
        children of fragment nodes
        :return: 2-tuple: ffs, conserved_val
          ffs = an array of FragmentFlow records reporting the traversal, beginning with self
          conserved_val = the magnitude of the flow with respect to the conserved quantity, if applicable (or None)
        """

        # first check for cycles
        if frags_seen is None:
            frags_seen = set()

        if self.reference_entity is None:
            if self.uuid in frags_seen:
                # this should really get resolved into a loop-closing algorithm
                raise InvalidParentChild('Frag %s seeing self\n %s' % (self.uuid, '; '.join(frags_seen)))
            frags_seen.add(self.uuid)

        if _balance is None:
            _scen_ev = self._match_scenario_ev(scenarios)
            ev = self.exchange_value(_scen_ev)
        else:
            _scen_ev = None
            self.dbg_print('%g balance' % _balance, level=2)
            ev = _balance
            '''
            if self._check_observability(scenario):
                self._cache_balance_ev(_balance, scenario, observed)
            '''

        magnitude = upstream_nw * ev
        _scen_term = self._match_scenario_term(scenarios)
        term = self._terminations[_scen_term]

        if self.reference_entity is None:
            node_weight = upstream_nw
            if magnitude == 0:
                self.dbg_print('Unobserved reference fragment')
                node_weight = 0
        else:
            node_weight = magnitude

        f_conv = term.node_weight_multiplier

        node_weight *= f_conv

        self.dbg_print('magnitude: %g f_conv: %g node_weight: %g' % (magnitude, f_conv, node_weight))

        conserved_val = None
        if conserved_qty is not None:
            if self.is_balance:
                self.dbg_print('Found balance flow')
                raise FoundBalanceFlow  # to be caught
            cf = conserved_qty.cf(self.flow)
            self.dbg_print('consrv cf %g for qty %s' % (cf, conserved_qty), level=3)
            conserved_val = ev * cf
            if conserved_val == 0:
                conserved = False
            else:
                conserved = True
                if self.direction == 'Output':  # convention: inputs to parent are positive
                    conserved_val *= -1
                self.dbg_print('conserved_val %g' % conserved_val, level=2)
        elif self.is_balance:
            # traversing balance flow after FoundBalanceFlow exception
            conserved = True
        else:
            conserved = self.conserved

        # print('%6f %6f %s' % (magnitude, node_weight, self))
        # this is the only place a FragmentFlow is created
        # TODO: figure out how to cache + propagate matched scenarios ... in progress
        ff = FragmentFlow(self, magnitude, node_weight, term, conserved, match_ev=_scen_ev, match_term=_scen_term,
                          flow_conversion=f_conv)

        '''
        now looking forward: is our termination a cutoff, background, foreground or subfragment?
        '''
        if magnitude == 0:
            # no flow to follow
            self.dbg_print('zero magnitude')
            return [ff], conserved_val
        elif term.is_null or term.is_context:
            # cutoff /context and background end traversal
            self.dbg_print('cutoff or bg')
            return [ff], conserved_val

        if term.is_fg or term.is_process or not term.valid:
            self.dbg_print('fg')
            ffs = self._traverse_fg_node(ff, scenarios, frags_seen)

        else:
            self.dbg_print('subfrag')
            ffs = self._traverse_subfragment(ff, scenarios, frags_seen)

        return ffs, conserved_val


def _do_subfragment_traversal(ff, scenarios, frags_seen):
    """
    This turns out to be surprisingly complicated. So we now have:
     - LcFragment._traverse_node <-- which is called recursively
      + scenario matching + finds ev and term
      - selects handler based on term type:
       - LcFragment._subfragment_traversal
       |- invokes (static) _do_subfragment_traversal (YOU ARE HERE)
       ||- calls [internally recursive] term_node.unit_flows, which is just a wrapper for
       || - (static) group_ios
       ||  + reference flow and autoconsumption handling
       || /
       ||/
       |+ reference flow matching and normalizing
       + child flow matching and further recursion --> into LcFragment._traverse_node
      /
     /
     ffs, conserved_val


     - (static) group_ios
     - nested inside LcFragment.unit_flows, which is really just a wrapper
     - called from
     - called from

    :param ff: The FragmentFlow whose subfragment is being traversed
    :param scenarios: to pass to subfragment
    :return: ffs [subfragment traversal appended], unit_inv [with reference flow removed], downstream node weight
    """
    term = ff.term  # note that we ignore term.direction in subfragment traversal
    node_weight = ff.node_weight
    self = ff.fragment

    unit_inv, subfrags = term.term_node.unit_flows(scenario=scenarios, frags_seen=frags_seen)

    # find the inventory flow that matches us
    # use term_flow over term_node.flow because that allows client code to specify inverse traversal knowing
    #  only the sought flow.
    # unit_flows guarantees that there is exactly one of these flows (except in the case of cumulating flows!
    # see group_ios)
    try:
        match = next(k for k in unit_inv if k.fragment.flow == term.term_flow)
    except StopIteration:
        print('Flow mismatch Traversing:\n%s' % self)
        print('Term flow: %s' % term.term_flow.link)
        print(term.serialize())
        for k in unit_inv:
            print('%s' % k.fragment.flow.link)
        raise MissingFlow('Term flow: %s' % term.term_flow.link)
    self.dbg_print('matched flow %s' % match.fragment.flow.link)

    unit_inv.remove(match)

    in_ex = match.magnitude  # this is the inbound exchange value for the driven fragment
    if in_ex == 0:
        # this indicates a non-consumptive pass-thru fragment.
        print('Frag %.5s: Zero inbound exchange' % self.uuid)
        raise ZeroInboundExchange
    if match.fragment.direction == self.direction:  # match direction is w.r.t. subfragment
        # self is driving subfragment in reverse
        self.dbg_print('reverse-driven subfragment %.3s' % match.fragment.uuid)
        in_ex *= -1

    # node weight for the driven [downstream] fragment
    downstream_nw = node_weight / in_ex

    '''
    # then we add the results of the subfragment, either in aggregated or disaggregated form
    if term.descend:
        # if appending, we are traversing in situ, so do scale
        self.dbg_print('descending', level=0)
        for i in subfrags:
            i.scale(downstream_nw)
        ffs = [ff]
        ffs.extend(subfrags)
    else:
        # if aggregating, we are only setting unit scores- so don't scale
        self.dbg_print('aggregating', level=0)
        ff.aggregate_subfragments(subfrags, scenarios=scenarios)  # include params to reproduce
        ff.node_weight = downstream_nw  # NOTE: this may be problematic; undone in lca_disclosures
        ffs = [ff]
    '''
    # now, we abolish descend as a traversal parameter and make it only an LCIA parameter
    # therefore, we retain subfragments always, we just have to decide how to scale them
    self.dbg_print('%d subfrags resulting from this traversal' % len(subfrags), 1)
    ff.aggregate_subfragments(subfrags, scenarios=scenarios)
    ff.node_weight = downstream_nw
    ffs = [ff]

    return ffs, unit_inv, downstream_nw




from ...implementations import AntelopeForegroundImplementation
from antelope_core.lcia_results import LciaResult
from ...refs.fragment_ref import ParentFragment


class AntelopeV1ForegroundImplementation(AntelopeForegroundImplementation):
    """
    Overrides the default implementation to handle the AntelopeV1 case
    """

    '''
    Implementation Method Overrides
    Unimplemented overrides will simply fallthrough to default:

    def exchanges(self, process, **kwargs):
        pass

    def exchange_values(self, process, flow, direction, termination=None, **kwargs):
        pass

    def inventory(self, process, ref_flow=None, **kwargs):
        pass
    '''
    def top(self, fragment, **kwargs):
        """
        a neutered top() here just returns self
        :param fragment:
        :param kwargs:
        :return:
        """
        if hasattr(fragment, 'external_ref'):
            fragment = fragment.external_ref
        return self._archive.retrieve_or_fetch_entity(fragment)

    def get_reference(self, key):
        raise ParentFragment

    def traverse(self, fragment, scenario=None, **kwargs):
        if hasattr(fragment, 'external_ref'):
            fragment = fragment.external_ref
        if scenario is not None:
            endpoint = 'scenarios/%s/%s/fragmentflows' % (scenario, fragment)
        else:
            endpoint = '%s/fragmentflows' % fragment

        self._archive.fetch_flows(fragment)

        ffs = self._archive.get_endpoint(endpoint)
        for ff in ffs:
            if 'fragmentStageID' in ff:
                ff['StageName'] = self._archive.get_stage_name(ff['fragmentStageID'])
        return [self._archive.make_fragment_flow(ff)
                for ff in sorted(ffs, key=lambda x: ('parentFragmentFlowID' in x, x['fragmentFlowID']))]

    def fragment_lcia(self, fragment, quantity_ref, scenario=None, refresh=False, **kwargs):
        if scenario is None:
            scenario = '1'

        if hasattr(fragment, 'external_ref'):
            fragment = fragment.external_ref

        qi = self._archive.make_interface('quantity')
        lcia_q = qi.get_lcia_quantity(quantity_ref)
        endpoint = 'scenarios/%s/%s/%s/lciaresults' % (scenario, fragment, lcia_q.external_ref)
        lcia_r = self._archive.get_endpoint(endpoint, cache=False)
        if lcia_r is None or (isinstance(lcia_r, list) and all(i is None for i in lcia_r)):
            res = LciaResult(lcia_q, scenario=scenario)
            return res

        res = LciaResult(lcia_q, scenario=lcia_r.pop('scenarioID'))
        total = lcia_r.pop('total')

        for component in lcia_r['lciaScore']:
            qi.add_lcia_component(res, component)

        qi.check_total(res.total(), total)

        return res





import unittest

from antelope import comp_dir
from antelope_core.entities.tests import BasicEntityTest
from ..entities.fragment_editor import create_fragment
from ..terminations import FlowTermination


class FlowTerminationTestCase(BasicEntityTest):
    """
    Mainly want to test initialization (for process ref and flow [context]), enumeration of "unobserved exchanges",
    and LCIA
    """
    '''
    _qdb = None

    @property
    def qdb(self):
        if self._qdb is None:
            self._qdb = Qdb()
            q = next(self.A.search('quantity', Name='coolness'))
            qi = self.A.make_interface('quantity')
            for cf in qi.factors(q):
                self._qdb.add_cf(cf)
        return self._qdb
    '''

    def _petro_frag(self):
        rx = next(x for x in self.petro.references() if x.flow['Name'].startswith('Diesel'))
        return create_fragment(rx.flow, rx.direction, origin='test.termination')

    def _petro_term(self):
        """
        NOTE: this term will have a phony / proxy background implementation, since we don't have a true catalog query
        :return:
        """
        frag = self._petro_frag()
        return frag.terminate(self.petro.make_ref(self.A.query), term_flow=frag.flow)

    def test_create_term(self):
        frag = self._petro_frag()
        term = FlowTermination(frag, self.petro, term_flow=frag.flow)
        self.assertIs(frag.flow, term.term_flow)
        self.assertEqual(comp_dir(frag.direction), term.direction)
        self.assertTrue(term.is_process)
        self.assertFalse(term.is_null)
        self.assertFalse(term.is_null)
        self.assertFalse(term.is_subfrag)
        self.assertTrue(term.is_bg)  # bg now means "no child flows"
        self.assertFalse(term.term_is_bg)

    def test_unobserved(self):
        term = self._petro_term()
        ux = [x for x in term.unobserved_exchanges()]
        self.assertEqual(len(ux), 42)

    def _frag_with_child(self):
        frag = self._petro_frag()
        frag.terminate(self.petro.make_ref(self.A.query), term_flow=frag.flow)
        lead = next(x for x in self.petro.inventory(frag.flow) if x.flow['Name'].startswith('Lead'))
        c = create_fragment(flow=lead.flow, direction=lead.direction, parent=frag, value=lead.value)
        c.terminate(self.A.tm[lead.flow.context])
        return frag

    def test_unobserved_with_child(self):
        frag = self._frag_with_child()
        ux = [x for x in frag.term.unobserved_exchanges()]
        self.assertEqual(len(ux), 41)

    def _get_coolness(self):
        q = next(self.A.search('quantity', Name='coolness'))
        return q.make_ref(self.A.query)

    def test_lcia(self):
        term = self._petro_term()
        q = self._get_coolness()
        res = term.compute_unit_score(q)
        self.assertAlmostEqual(res.total(), 0.00013695)

    def test_lcia_with_child(self):
        frag = self._frag_with_child()
        q = self._get_coolness()
        res = frag.term.compute_unit_score(q)
        self.assertAlmostEqual(res.total(), 0.00012425)

    def test_fg(self):
        frag = self._frag_with_child()
        z = next(frag.child_flows)
        self.assertFalse(z.term.is_fg)
        self.assertTrue(z.term.is_context)
        self.assertIs(z.flow, z.term.term_flow)
        self.assertEqual(z.direction, comp_dir(z.term.direction))

    def test_fg_lcia(self):
        frag = self._frag_with_child()
        z = next(frag.child_flows)
        q = self._get_coolness()
        res = z.term.compute_unit_score(q)
        self.assertEqual(res.total(), q.cf(z.flow))

    def test_traversal(self):
        c = self._frag_with_child()
        c.observe(accept_all=True)  # fragment_lcia only operates on observed exchanges
        q = self._get_coolness()
        res1 = q.do_lcia(self.petro.inventory(c.flow))
        res2 = c.fragment_lcia(q)
        self.assertAlmostEqual(res1.total(), res2.total(), places=15)

if __name__ == '__main__':
    unittest.main()
